#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest
import numpy as np
from typing import List, Union, Optional, Tuple

from nomad.units import ureg
from nomad.datamodel import EntryArchive

from nomad_simulations.schema_packages.atoms_state import (
    OrbitalsState,
    CoreHole,
    HubbardInteractions,
    AtomsState,
)

from . import logger


class TestOrbitalsState:
    """
    Test the `OrbitalsState` class defined in atoms_state.py.
    """

    @staticmethod
    def add_state(
        orbital_state: OrbitalsState,
        l_number: int,
        ml_number: Optional[int],
        ms_number: Optional[float],
        j_number: Optional[List[float]],
        mj_number: Optional[List[float]],
    ) -> None:
        """Adds l and ml quantum numbers to the `OrbitalsState` section."""
        orbital_state.l_quantum_number = l_number
        orbital_state.ml_quantum_number = ml_number
        orbital_state.ms_quantum_number = ms_number
        orbital_state.j_quantum_number = j_number
        orbital_state.mj_quantum_number = mj_number

    @pytest.mark.parametrize(
        'number_label, values, results',
        [
            ('n_quantum_number', [-1, 0, 1, 2], [False, False, True, True]),
            ('l_quantum_number', [-2, 0, 1, 2], [False, False, True, True]),
            # l_quantum_number == 2 when testing 'ml_quantum_number'
            ('ml_quantum_number', [-3, 5, -2, 1], [False, False, True, True]),
            ('ms_quantum_number', [0, 10, -0.5, 0.5], [False, False, True, True]),
        ],
    )
    def test_validate_quantum_numbers(
        self, number_label: str, values: List[int], results: List[bool]
    ):
        """
        Test the `validate_quantum_numbers` method.

        Args:
            number_label (str): The quantum number string to be tested.
            values (List[int]): The values stored in `OrbitalState`.
            results (List[bool]): The expected results after validation.
        """
        orbital_state = OrbitalsState(n_quantum_number=2)
        for val, res in zip(values, results):
            if number_label == 'ml_quantum_number':
                orbital_state.l_quantum_number = 2
            setattr(orbital_state, number_label, val)
            assert orbital_state.validate_quantum_numbers(logger=logger) == res

    @pytest.mark.parametrize(
        'quantum_name, value, expected_result',
        [
            ('l', 0, 's'),
            ('l', 1, 'p'),
            ('l', 2, 'd'),
            ('l', 3, 'f'),
            ('l', 4, None),
            ('ml', -1, 'x'),
            ('ml', 0, 'z'),
            ('ml', 1, 'y'),
            ('ml', -2, None),
            ('ms', -0.5, 'down'),
            ('ms', 0.5, 'up'),
            ('ms', -0.75, None),
            ('no_attribute', None, None),
        ],
    )
    def test_number_and_symbol(
        self,
        quantum_name: str,
        value: Union[int, float],
        expected_result: Optional[str],
    ):
        """
        Test the number and symbol resolution for each of the quantum numbers defined in the parametrization.

        Args:
            quantum_name (str): The quantum number string to be tested.
            value (Union[int, float]): The value stored in `OrbitalState`.
            expected_result (Optional[str]): The expected result after resolving the counter-type.
        """
        # Adding quantum numbers to the `OrbitalsState` section
        orbital_state = OrbitalsState(n_quantum_number=2)
        if quantum_name == 'ml':  # l_quantum_number must be specified
            orbital_state.l_quantum_number = 1
        setattr(orbital_state, f'{quantum_name}_quantum_number', value)

        # Making sure that the `'number'` is assigned
        resolved_type = orbital_state.resolve_number_and_symbol(
            quantum_name=quantum_name, quantum_type='number', logger=logger
        )
        assert resolved_type == value

        # Resolving if the counter-type is assigned
        resolved_countertype = orbital_state.resolve_number_and_symbol(
            quantum_name=quantum_name, quantum_type='symbol', logger=logger
        )
        assert resolved_countertype == expected_result

    @pytest.mark.parametrize(
        'l_quantum_number, ml_quantum_number, j_quantum_number, mj_quantum_number, ms_quantum_number, degeneracy',
        [
            (1, None, None, None, 0.5, 3),
            (1, None, None, None, None, 6),
            (1, -1, None, None, 0.5, 1),
            (1, -1, None, None, None, 2),
            # ! `RusselSanders` uses the electronic state, but here we are defining single orbitals. We need new methodology to treat these cases separately
            # ! and these tests for j_quantum_number and mj_quantum_number need to be updated.
            (1, -1, [1 / 2, 3 / 2], None, None, 6),
            (1, -1, [1 / 2, 3 / 2], [-3 / 2, 1 / 2, 1 / 2, 3 / 2], None, 2),
        ],
    )
    def test_degeneracy(
        self,
        l_quantum_number: int,
        ml_quantum_number: Optional[int],
        j_quantum_number: Optional[List[float]],
        mj_quantum_number: Optional[List[float]],
        ms_quantum_number: Optional[float],
        degeneracy: int,
    ):
        """
        Test the degeneracy of each orbital states defined in the parametrization.

        Args:
            l_quantum_number (int): The angular momentum quantum number.
            ml_quantum_number (Optional[int]): The magnetic quantum number.
            j_quantum_number (Optional[List[float]]): The total angular momentum quantum number.
            mj_quantum_number (Optional[List[float]]): The magnetic quantum number for the total angular momentum.
            ms_quantum_number (Optional[float]): The spin quantum number.
            degeneracy (int): The expected degeneracy of the orbital state.
        """
        orbital_state = OrbitalsState(n_quantum_number=2)
        self.add_state(
            orbital_state,
            l_quantum_number,
            ml_quantum_number,
            ms_quantum_number,
            j_quantum_number,
            mj_quantum_number,
        )
        assert orbital_state.resolve_degeneracy() == degeneracy

    def test_normalize(self):
        """
        Test the normalization of the `OrbitalsState`. Inputs are defined as the quantities of the `OrbitalsState` section.
        """
        orbital_state = OrbitalsState(n_quantum_number=2)
        self.add_state(orbital_state, 2, -2, None, None, None)
        orbital_state.normalize(EntryArchive(), logger)
        assert orbital_state.n_quantum_number == 2
        assert orbital_state.l_quantum_number == 2
        assert orbital_state.l_quantum_symbol == 'd'
        assert orbital_state.ml_quantum_number == -2
        assert orbital_state.ml_quantum_symbol == 'xy'
        assert orbital_state.degeneracy == 2


class TestCoreHole:
    """
    Test the `CoreHole` class defined in atoms_state.py.
    """

    @pytest.mark.parametrize(
        'orbital_ref, degeneracy, n_excited_electrons, occupation',
        [
            (OrbitalsState(l_quantum_number=1), 6, 0.5, 5.5),
            (OrbitalsState(l_quantum_number=1, ml_quantum_number=-1), 2, 0.5, 1.5),
            (None, None, 0.5, None),
        ],
    )
    def test_occupation(
        self,
        orbital_ref: Optional[OrbitalsState],
        degeneracy: Optional[int],
        n_excited_electrons: float,
        occupation: Optional[float],
    ):
        """
        Test the occupation of a core hole for a given set of orbital reference and degeneracy.

        Args:
            orbital_ref (Optional[OrbitalsState]): The orbital reference of the core hole.
            degeneracy (Optional[int]): The degeneracy of the orbital reference.
            n_excited_electrons (float): The number of excited electrons.
            occupation (Optional[float]): The expected occupation of the core hole.
        """
        core_hole = CoreHole(
            orbital_ref=orbital_ref, n_excited_electrons=n_excited_electrons
        )
        if orbital_ref is not None:
            assert orbital_ref.resolve_degeneracy() == degeneracy
        resolved_occupation = core_hole.resolve_occupation(logger=logger)
        if resolved_occupation is not None:
            assert np.isclose(resolved_occupation, occupation)
        else:
            assert resolved_occupation == occupation

    @pytest.mark.parametrize(
        'orbital_ref, n_excited_electrons, dscf_state, results',
        [
            (OrbitalsState(l_quantum_number=1), -0.5, None, (-0.5, None, None)),
            (OrbitalsState(l_quantum_number=1), 0.5, None, (0.5, None, 5.5)),
            (
                OrbitalsState(l_quantum_number=1, ml_quantum_number=-1),
                0.5,
                None,
                (0.5, None, 1.5),
            ),
            (OrbitalsState(l_quantum_number=1), 0.5, 'initial', (None, 1, None)),
            (OrbitalsState(l_quantum_number=1), 0.5, 'final', (0.5, None, 5.5)),
            (None, 0.5, None, (0.5, None, None)),
        ],
    )
    def test_normalize(
        self,
        orbital_ref: Optional[OrbitalsState],
        n_excited_electrons: Optional[float],
        dscf_state: Optional[str],
        results: Tuple[Optional[float], Optional[float], Optional[float]],
    ):
        """
        Test the normalization of the `CoreHole`. Inputs are defined as the quantities of the `CoreHole` section.

        Args:
            orbital_ref (Optional[OrbitalsState]): The orbital reference of the core hole.
            n_excited_electrons (Optional[float]): The number of excited electrons.
            dscf_state (Optional[str]): The DSCF state of the core hole.
            results (Tuple[Optional[float], Optional[float], Optional[float]]): The expected results after normalization.
        """
        core_hole = CoreHole(
            orbital_ref=orbital_ref,
            n_excited_electrons=n_excited_electrons,
            dscf_state=dscf_state,
        )
        core_hole.normalize(EntryArchive(), logger)
        assert core_hole.n_excited_electrons == results[0]
        if core_hole.orbital_ref:
            assert core_hole.orbital_ref.degeneracy == results[1]
            assert core_hole.orbital_ref.occupation == results[2]


class TestHubbardInteractions:
    """
    Test the `HubbardInteractions` class defined in atoms_state.py.
    """

    @pytest.mark.parametrize(
        'slater_integrals, results',
        [
            ([3.0, 2.0, 1.0], (0.1429146, -0.0357286, 0.0893216)),
            (None, (None, None, None)),
            ([3.0, 2.0, 1.0, 0.5], (None, None, None)),
        ],
    )
    def test_u_interactions(
        self,
        slater_integrals: Optional[List[float]],
        results: Tuple[Optional[float], Optional[float], Optional[float]],
    ):
        """
        Test the Hubbard interactions `U`, `U'`, and `J` for a given set of Slater integrals.

        Args:
            slater_integrals (Optional[List[float]]): The Slater integrals of the Hubbard interactions.
            results (Tuple[Optional[float], Optional[float], Optional[float]]): The expected results of the Hubbard interactions.
        """
        # Adding `slater_integrals` to the `HubbardInteractions` section
        hubbard_interactions = HubbardInteractions()
        if slater_integrals is not None:
            hubbard_interactions.slater_integrals = slater_integrals * ureg.eV

        # Resolving U, U', and J from class method
        (
            u_interaction,
            u_interorbital_interaction,
            j_hunds_coupling,
        ) = hubbard_interactions.resolve_u_interactions(logger=logger)

        if None not in (u_interaction, u_interorbital_interaction, j_hunds_coupling):
            assert np.isclose(u_interaction.to('eV').magnitude, results[0])
            assert np.isclose(u_interorbital_interaction.to('eV').magnitude, results[1])
            assert np.isclose(j_hunds_coupling.to('eV').magnitude, results[2])
        else:
            assert (
                u_interaction,
                u_interorbital_interaction,
                j_hunds_coupling,
            ) == results

    @pytest.mark.parametrize(
        'u_interaction, j_local_exchange_interaction, u_effective',
        [
            (3.0, 1.0, 2.0),
            (-3.0, 1.0, None),
            (3.0, None, 3.0),
            (None, 1.0, None),
        ],
    )
    def test_u_effective(
        self,
        u_interaction: Optional[float],
        j_local_exchange_interaction: Optional[float],
        u_effective: Optional[float],
    ):
        """
        Test the effective Hubbard interaction `U_eff` for a given set of Hubbard interactions `U` and `J`.

        Args:
            u_interaction (Optional[float]): The Hubbard interaction `U`.
            j_local_exchange_interaction (Optional[float]): The Hubbard interaction `J`.
            u_effective (Optional[float]): The expected effective Hubbard interaction `U_eff`.
        """
        # Adding `u_interaction` and `j_local_exchange_interaction` to the `HubbardInteractions` section
        hubbard_interactions = HubbardInteractions()
        if u_interaction is not None:
            hubbard_interactions.u_interaction = u_interaction * ureg.eV
        if j_local_exchange_interaction is not None:
            hubbard_interactions.j_local_exchange_interaction = (
                j_local_exchange_interaction * ureg.eV
            )

        # Resolving Ueff from class method
        resolved_u_effective = hubbard_interactions.resolve_u_effective(logger=logger)
        if resolved_u_effective is not None:
            assert np.isclose(resolved_u_effective.to('eV').magnitude, u_effective)
        else:
            assert resolved_u_effective == u_effective

    def test_normalize(self):
        """
        Test the normalization of the `HubbardInteractions`. Inputs are defined as the quantities of the `HubbardInteractions` section.
        """
        # ? Is this enough for testing? Can we do more?
        hubbard_interactions = HubbardInteractions(
            u_interaction=3.0 * ureg.eV,
            u_interorbital_interaction=1.0 * ureg.eV,
            j_hunds_coupling=2.0 * ureg.eV,
            j_local_exchange_interaction=2.0 * ureg.eV,
        )
        hubbard_interactions.normalize(EntryArchive(), logger)
        assert np.isclose(hubbard_interactions.u_effective.to('eV').magnitude, 1.0)
        assert np.isclose(hubbard_interactions.u_interaction.to('eV').magnitude, 3.0)


class TestAtomsState:
    """
    Tests the `AtomsState` class defined in atoms_state.py.
    """

    @pytest.mark.parametrize(
        'chemical_symbol, atomic_number',
        [
            ('Fe', 26),
            ('H', 1),
            ('Cu', 29),
            ('O', 8),
        ],
    )
    def test_chemical_symbol_and_atomic_number(
        self, chemical_symbol: str, atomic_number: int
    ):
        """
        Test the `chemical_symbol` and `atomic_number` resolution for the `AtomsState` section.

        Args:
            chemical_symbol (str): The chemical symbol of the atom.
            atomic_number (int): The atomic number of the atom.
        """
        # Testing `chemical_symbol`
        atom_state = AtomsState(chemical_symbol=chemical_symbol)
        assert atom_state.resolve_atomic_number(logger=logger) == atomic_number
        # Testing `atomic_number`
        atom_state.atomic_number = atomic_number
        assert atom_state.resolve_chemical_symbol(logger=logger) == chemical_symbol



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import TYPE_CHECKING, List

import numpy as np

from nomad.config import config
from nomad.datamodel.data import EntryData
from nomad.datamodel.metainfo.annotations import ELNAnnotation
from nomad.datamodel.metainfo.basesections import Activity, Entity
from nomad.metainfo import Datetime, Quantity, SchemaPackage, Section, SubSection

if TYPE_CHECKING:
    from nomad.datamodel.datamodel import EntryArchive
    from structlog.stdlib import BoundLogger

from nomad_simulations.schema_packages.model_method import ModelMethod
from nomad_simulations.schema_packages.model_system import ModelSystem
from nomad_simulations.schema_packages.outputs import Outputs
from nomad_simulations.schema_packages.utils import (
    get_composition,
    is_not_representative,
)

configuration = config.get_plugin_entry_point(
    'nomad_simulations.schema_packages:nomad_simulations_plugin'
)

m_package = SchemaPackage()


class Program(Entity):
    """
    A base section used to specify a well-defined program used for computation.

    Synonyms:
     - code
     - software
    """

    name = Quantity(
        type=str,
        description="""
        The name of the program.
        """,
        a_eln=ELNAnnotation(component='StringEditQuantity'),
    )

    version = Quantity(
        type=str,
        description="""
        The version label of the program.
        """,
        a_eln=ELNAnnotation(component='StringEditQuantity'),
    )

    link = Quantity(
        type=str,
        description="""
        Website link to the program in published information.
        """,
        a_eln=ELNAnnotation(component='URLEditQuantity'),
    )

    version_internal = Quantity(
        type=str,
        description="""
        Specifies a program version tag used internally for development purposes.
        Any kind of tagging system is supported, including git commit hashes.
        """,
        a_eln=ELNAnnotation(component='StringEditQuantity'),
    )

    compilation_host = Quantity(
        type=str,
        description="""
        Specifies the host on which the program was compiled.
        """,
        a_eln=ELNAnnotation(component='StringEditQuantity'),
    )

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        pass


class BaseSimulation(Activity):
    """
    A computational simulation that produces output data from a given input model system
    and input methodological parameters.

    Synonyms:
        - computation
        - calculation
    """

    m_def = Section(
        links=['https://liusemweb.github.io/mdo/core/1.1/index.html#Calculation']
    )

    datetime_end = Quantity(
        type=Datetime,
        description="""
        The date and time when this computation ended.
        """,
        a_eln=ELNAnnotation(component='DateTimeEditQuantity'),
    )

    cpu1_start = Quantity(
        type=np.float64,
        unit='second',
        description="""
        The starting time of the computation on the (first) CPU 1.
        """,
        a_eln=ELNAnnotation(component='NumberEditQuantity'),
    )

    cpu1_end = Quantity(
        type=np.float64,
        unit='second',
        description="""
        The end time of the computation on the (first) CPU 1.
        """,
        a_eln=ELNAnnotation(component='NumberEditQuantity'),
    )

    wall_start = Quantity(
        type=np.float64,
        unit='second',
        description="""
        The internal wall-clock time from the starting of the computation.
        """,
        a_eln=ELNAnnotation(component='NumberEditQuantity'),
    )

    wall_end = Quantity(
        type=np.float64,
        unit='second',
        description="""
        The internal wall-clock time from the end of the computation.
        """,
        a_eln=ELNAnnotation(component='NumberEditQuantity'),
    )

    program = SubSection(sub_section=Program.m_def, repeats=False)

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        pass


class Simulation(BaseSimulation, EntryData):
    """
    A `Simulation` is a computational calculation that produces output data from a given input model system
    and input (model) methodological parameters. The output properties obtained from the simulation are stored
    in a list under `outputs`.

    Each sub-section of `Simulation` is defined in their corresponding modules: `model_system.py`, `model_method.py`,
    and `outputs.py`.

    The basic entry data for a `Simulation`, known as `SinglePoint` workflow, contains all the self-consistent (SCF) steps
    performed to converge the calculation, i.e., we do not split each SCF step in its own entry but rather group them in a general one.

    Synonyms:
        - calculation
        - computation
    """

    model_system = SubSection(sub_section=ModelSystem.m_def, repeats=True)

    model_method = SubSection(sub_section=ModelMethod.m_def, repeats=True)

    outputs = SubSection(sub_section=Outputs.m_def, repeats=True)

    def _set_system_branch_depth(
        self, system_parent: ModelSystem, branch_depth: int = 0
    ):
        for system_child in system_parent.model_system:
            system_child.branch_depth = branch_depth + 1
            self._set_system_branch_depth(
                system_parent=system_child, branch_depth=branch_depth + 1
            )

    def resolve_composition_formula(self, system_parent: ModelSystem) -> None:
        """Determine and set the composition formula for `system_parent` and all of its
        descendants.

        Args:
            system_parent (ModelSystem): The upper-most level of the system hierarchy to consider.
        """

        def set_composition_formula(
            system: ModelSystem, subsystems: List[ModelSystem], atom_labels: List[str]
        ) -> None:
            """Determine the composition formula for `system` based on its `subsystems`.
            If `system` has no children, the atom_labels are used to determine the formula.

            Args:
                system (ModelSystem): The system under consideration.
                subsystems (List[ModelSystem]): The children of system.
                atom_labels (List[str]): The global list of atom labels corresponding
                to the atom indices stored in system.
            """
            if not subsystems:
                atom_indices = (
                    system.atom_indices if system.atom_indices is not None else []
                )
                subsystem_labels = (
                    [np.array(atom_labels)[atom_indices]]
                    if atom_labels
                    else ['Unknown' for atom in range(len(atom_indices))]
                )
            else:
                subsystem_labels = [
                    subsystem.branch_label
                    if subsystem.branch_label is not None
                    else 'Unknown'
                    for subsystem in subsystems
                ]
            if system.composition_formula is None:
                system.composition_formula = get_composition(
                    children_names=subsystem_labels
                )

        def get_composition_recurs(system: ModelSystem, atom_labels: List[str]) -> None:
            """Traverse the system hierarchy downward and set the branch composition for
            all (sub)systems at each level.

            Args:
                system (ModelSystem): The system to traverse downward.
                atom_labels (List[str]): The global list of atom labels corresponding
                to the atom indices stored in system.
            """
            subsystems = system.model_system
            set_composition_formula(
                system=system, subsystems=subsystems, atom_labels=atom_labels
            )
            if subsystems:
                for subsystem in subsystems:
                    get_composition_recurs(system=subsystem, atom_labels=atom_labels)

        atoms_state = (
            system_parent.cell[0].atoms_state if system_parent.cell is not None else []
        )
        atom_labels = (
            [atom.chemical_symbol for atom in atoms_state]
            if atoms_state is not None
            else []
        )
        get_composition_recurs(system=system_parent, atom_labels=atom_labels)

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super(EntryData, self).normalize(archive, logger)

        # Finding which is the representative system of a calculation: typically, we will
        # define it as the last system reported (TODO CHECK THIS!).
        # TODO extend adding the proper representative system extraction using `normalizer.py`
        if self.model_system is None:
            logger.error('No system information reported.')
            return
        system_ref = self.model_system[-1]
        # * We define is_representative in the parser
        # system_ref.is_representative = True
        self.m_cache['system_ref'] = system_ref

        # Setting up the `branch_depth` in the parent-child tree
        for system_parent in self.model_system:
            system_parent.branch_depth = 0
            if len(system_parent.model_system) == 0:
                continue
            self._set_system_branch_depth(system_parent=system_parent)

            if is_not_representative(model_system=system_parent, logger=logger):
                continue
            self.resolve_composition_formula(system_parent=system_parent)


m_package.__init_metainfo__()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import re
from typing import TYPE_CHECKING, List, Optional

import numpy as np

from nomad.datamodel.data import ArchiveSection
from nomad.datamodel.metainfo.annotations import ELNAnnotation
from nomad.metainfo import MEnum, Quantity, SubSection

if TYPE_CHECKING:
    from nomad.metainfo import Section, Context
    from nomad.datamodel.datamodel import EntryArchive
    from structlog.stdlib import BoundLogger

from nomad_simulations.schema_packages.atoms_state import CoreHole, OrbitalsState
from nomad_simulations.schema_packages.model_system import ModelSystem
from nomad_simulations.schema_packages.numerical_settings import NumericalSettings
from nomad_simulations.schema_packages.utils import is_not_representative


class ModelMethod(ArchiveSection):
    """
    Model method input parameters and numerical settings used to simulate materials properties.

    # ! add more description once section is finished
    """

    normalizer_level = 1

    name = Quantity(
        type=str,
        description="""
        Name of the model method. This is typically used to easy identification of the `ModelMethod` section.

        Suggested values: 'DFT', 'TB', 'GE', 'BSE', 'DMFT', 'NMR', 'kMC'.
        """,
        a_eln=ELNAnnotation(component='StringEditQuantity'),
    )

    type = Quantity(
        type=str,
        description="""
        Identifier used to further specify the type of model Hamiltonian. Example: a TB
        model can be 'Wannier', 'DFTB', 'xTB' or 'Slater-Koster'. This quantity should be
        rewritten to a MEnum when inheriting from this class.
        """,
        a_eln=ELNAnnotation(component='StringEditQuantity'),
    )

    external_reference = Quantity(
        type=str,
        description="""
        External reference to the model e.g. DOI, URL.
        """,
        a_eln=ELNAnnotation(component='URLEditQuantity'),
    )

    numerical_settings = SubSection(sub_section=NumericalSettings.m_def, repeats=True)

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


class ModelMethodElectronic(ModelMethod):
    """
    A base section used to define the parameters of a model Hamiltonian used in electronic structure
    calculations (TB, DFT, GW, BSE, DMFT, etc).
    """

    # ? Is this necessary or will it be defined in another way?
    is_spin_polarized = Quantity(
        type=bool,
        description="""
        If the simulation is done considering the spin degrees of freedom (then there are two spin
        channels, 'down' and 'up') or not.
        """,
    )

    # ? What about this quantity
    relativity_method = Quantity(
        type=MEnum(
            'scalar_relativistic',
            'pseudo_scalar_relativistic',
            'scalar_relativistic_atomic_ZORA',
        ),
        description="""
        Describes the relativistic treatment used for the calculation of the final energy
        and related quantities. If `None`, no relativistic treatment is applied.
        """,
        a_eln=ELNAnnotation(component='EnumEditQuantity'),
    )

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


class XCFunctional(ArchiveSection):
    """
    A base section used to define the parameters of an exchange or correlation functional.
    """

    libxc_name = Quantity(
        type=str,
        description="""
        Provides the name of one of the exchange or correlation (XC) functional following the libxc
        convention (see https://www.tddft.org/programs/libxc/).
        """,
        a_eln=ELNAnnotation(component='StringEditQuantity'),
    )

    name = Quantity(
        type=MEnum('exchange', 'correlation', 'hybrid', 'contribution'),
        description="""
        Name of the XC functional. It can be one of the following: 'exchange', 'correlation',
        'hybrid', or 'contribution'.
        """,
    )

    weight = Quantity(
        type=np.float64,
        description="""
        Weight of the functional. This quantity is relevant when defining linear combinations of the
        different functionals. If not specified, its value is 1.
        """,
        a_eln=ELNAnnotation(component='NumberEditQuantity'),
    )

    # ? add method to extract `name` from `libxc_name`

    def get_weight_name(self, weight: Optional[np.float64]) -> Optional[str]:
        """
        Returns the `weight` as a string with a "*" added at the end.

        Args:
            weight (Optional[np.float64]): The weight of the functional.

        Returns:
            (Optional[str]): The weight as a string with a "*" added at the end.
        """
        weight_name = ''
        if weight is not None:
            weight_name = f'{str(weight)}*'
        return weight_name

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)

        # Appending `weight` as a string to `libxc_name`
        libxc_name_weight = ''
        if self.weight is not None:
            libxc_name_weight = self.get_weight_name(self.weight)
        if '*' not in self.libxc_name:
            self.libxc_name = libxc_name_weight + self.libxc_name

        # ! check with @ndaelman-hu if this makes sense (COMMENTED OUT FOR NOW)
        # Appending `"+alpha"` in `libxc_name` for hybrids in which the `exact_exchange_mixing_factoris` included
        # libxc_name_alpha = ''
        # if (
        #     self.name == 'hybrid'
        #     and 'exact_exchange_mixing_factor' in self.parameters.keys()
        # ):
        #     libxc_name_alpha = f'+alpha'
        # if '+alpha' not in self.libxc_name:
        #     self.libxc_name = self.libxc_name + libxc_name_alpha


class DFT(ModelMethodElectronic):
    """
    A base section used to define the parameters used in a density functional theory (DFT) calculation.
    """

    # ? Do we need to define `type` for DFT+U?

    jacobs_ladder = Quantity(
        type=MEnum('LDA', 'GGA', 'metaGGA', 'hyperGGA', 'hybrid', 'unavailable'),
        description="""
        Functional classification in line with Jacob's Ladder. See:
            - https://doi.org/10.1063/1.1390175 (original paper)
            - https://doi.org/10.1103/PhysRevLett.91.146401 (meta-GGA)
            - https://doi.org/10.1063/1.1904565 (hyper-GGA)
        """,
    )

    xc_functionals = SubSection(sub_section=XCFunctional.m_def, repeats=True)

    exact_exchange_mixing_factor = Quantity(
        type=np.float64,
        description="""
        Amount of exact exchange mixed in with the XC functional (value range = [0, 1]).
        """,
        a_eln=ELNAnnotation(component='NumberEditQuantity'),
    )

    # ! MEnum this
    self_interaction_correction_method = Quantity(
        type=str,
        description="""
        Contains the name for the self-interaction correction (SIC) treatment used to
        calculate the final energy and related quantities. If skipped or empty, no special
        correction is applied.

        The following SIC methods are available:

        | SIC method                | Description                       |

        | ------------------------- | --------------------------------  |

        | `""`                      | No correction                     |

        | `"SIC_AD"`                | The average density correction    |

        | `"SIC_SOSEX"`             | Second order screened exchange    |

        | `"SIC_EXPLICIT_ORBITALS"` | (scaled) Perdew-Zunger correction explicitly on a
        set of orbitals |

        | `"SIC_MAURI_SPZ"`         | (scaled) Perdew-Zunger expression on the spin
        density / doublet unpaired orbital |

        | `"SIC_MAURI_US"`          | A (scaled) correction proposed by Mauri and co-
        workers on the spin density / doublet unpaired orbital |
        """,
        a_eln=ELNAnnotation(component='StringEditQuantity'),
    )

    van_der_waals_correction = Quantity(
        type=MEnum('TS', 'OBS', 'G06', 'JCHS', 'MDB', 'XC'),
        description="""
        Describes the Van der Waals (VdW) correction methodology. If `None`, no VdW correction is applied.

        | VdW method  | Reference                               |
        | --------------------- | ----------------------------------------- |
        | `"TS"`  | http://dx.doi.org/10.1103/PhysRevLett.102.073005 |
        | `"OBS"` | http://dx.doi.org/10.1103/PhysRevB.73.205101 |
        | `"G06"` | http://dx.doi.org/10.1002/jcc.20495 |
        | `"JCHS"` | http://dx.doi.org/10.1002/jcc.20570 |
        | `"MDB"` | http://dx.doi.org/10.1103/PhysRevLett.108.236402 and http://dx.doi.org/10.1063/1.4865104 |
        | `"XC"` | The method to calculate the VdW energy uses a non-local functional |
        """,
        a_eln=ELNAnnotation(component='EnumEditQuantity'),
    )

    def __init__(self, m_def: 'Section' = None, m_context: 'Context' = None, **kwargs):
        super().__init__(m_def, m_context, **kwargs)
        self._jacobs_ladder_map = {
            'lda': 'LDA',
            'gga': 'GGA',
            'mgg': 'meta-GGA',
            'hyb_mgg': 'hyper-GGA',
            'hyb': 'hybrid',
        }

    def resolve_libxc_names(
        self, xc_functionals: List[XCFunctional]
    ) -> Optional[List[str]]:
        """
        Resolves the `libxc_names` and sorts them from the list of `XCFunctional` sections.

        Args:
            xc_functionals (List[XCFunctional]): The list of `XCFunctional` sections.

        Returns:
            (Optional[List[str]]): The resolved and sorted `libxc_names`.
        """
        return sorted(
            [
                functional.libxc_name
                for functional in xc_functionals
                if functional.libxc_name is not None
            ]
        )

    def resolve_jacobs_ladder(
        self,
        libxc_names: List[str],
    ) -> str:
        """
        Resolves the `jacobs_ladder` from the `libxc_names`. The mapping (libxc -> NOMAD) is set in `self._jacobs_ladder_map`.

        Args:
            libxc_names (List[str]): The list of `libxc_names`.

        Returns:
            (str): The resolved `jacobs_ladder`.
        """
        if libxc_names is None:
            return 'unavailable'

        rung_order = {x: i for i, x in enumerate(self._jacobs_ladder_map.keys())}
        re_abbrev = re.compile(r'((HYB_)?[A-Z]{3})')

        abbrevs = []
        for xc_name in libxc_names:
            try:
                abbrev = re_abbrev.match(xc_name).group(1)
                abbrev = abbrev.lower() if abbrev == 'HYB_MGG' else abbrev[:3].lower()
                abbrevs.append(abbrev)
            except AttributeError:
                continue

        try:
            highest_rung_abbrev = max(abbrevs, key=lambda x: rung_order[x])
        except KeyError:
            return 'unavailable'
        return self._jacobs_ladder_map.get(highest_rung_abbrev, 'unavailable')

    def resolve_exact_exchange_mixing_factor(
        self, xc_functionals: List[XCFunctional], libxc_names: List[str]
    ) -> Optional[float]:
        """
        Resolves the `exact_exchange_mixing_factor` from the `xc_functionals` and `libxc_names`.

        Args:
            xc_functionals (List[XCFunctional]): The list of `XCFunctional` sections.
            libxc_names (List[str]): The list of `libxc_names`.

        Returns:
            (Optional[float]): The resolved `exact_exchange_mixing_factor`.
        """

        for functional in xc_functionals:
            if functional.name == 'hybrid':
                return functional.parameters.get('exact_exchange_mixing_factor')

        def _scan_patterns(patterns: List[str], xc_name: str) -> bool:
            return any(x for x in patterns if re.search('_' + x + '$', xc_name))

        for xc_name in libxc_names:
            if not re.search('_XC?_', xc_name):
                continue
            if re.search('_B3LYP[35]?$', xc_name):
                return 0.2
            elif _scan_patterns(['HSE', 'PBEH', 'PBE_MOL0', 'PBE_SOL0'], xc_name):
                return 0.25
            elif re.search('_M05$', xc_name):
                return 0.28
            elif re.search('_PBE0_13$', xc_name):
                return 1 / 3
            elif re.search('_PBE38$', xc_name):
                return 3 / 8
            elif re.search('_PBE50$', xc_name):
                return 0.5
            elif re.search('_M06_2X$', xc_name):
                return 0.54
            elif _scan_patterns(['M05_2X', 'PBE_2X'], xc_name):
                return 0.56
        return None

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)

        libxc_names = self.resolve_libxc_names(self.xc_functionals)
        if libxc_names is not None:
            # Resolves the `jacobs_ladder` from `libxc` mapping
            jacobs_ladder = self.resolve_jacobs_ladder(libxc_names)
            self.jacobs_ladder = (
                jacobs_ladder if self.jacobs_ladder is None else self.jacobs_ladder
            )

            # Resolves the `exact_exchange_mixing_factor` from the `xc_functionals` and `libxc_names`
            if self.xc_functionals is not None:
                exact_exchange_mixing_factor = (
                    self.resolve_exact_exchange_mixing_factor(
                        self.xc_functionals, libxc_names
                    )
                )
                self.exact_exchange_mixing_factor = (
                    exact_exchange_mixing_factor
                    if self.exact_exchange_mixing_factor is None
                    else self.exact_exchange_mixing_factor
                )


class TB(ModelMethodElectronic):
    """
    A base section containing the parameters pertaining to a tight-binding (TB) model calculation.
    The type of tight-binding model is specified in the `type` quantity.
    """

    type = Quantity(
        type=MEnum('DFTB', 'xTB', 'Wannier', 'SlaterKoster', 'unavailable'),
        default='unavailable',
        description="""
        Tight-binding model type.

        | Value | Reference |
        | --------- | ----------------------- |
        | `'DFTB'` | https://en.wikipedia.org/wiki/DFTB |
        | `'xTB'` | https://xtb-docs.readthedocs.io/en/latest/ |
        | `'Wannier'` | https://www.wanniertools.org/theory/tight-binding-model/ |
        | `'SlaterKoster'` | https://journals.aps.org/pr/abstract/10.1103/PhysRev.94.1498 |
        """,
        a_eln=ELNAnnotation(component='EnumEditQuantity'),
    )

    # ? these 2 quantities will change when `BasisSet` is defined
    n_orbitals = Quantity(
        type=np.int32,
        description="""
        Number of orbitals used as a basis to obtain the `TB` model.
        """,
    )

    orbitals_ref = Quantity(
        type=OrbitalsState,
        shape=['n_orbitals'],
        description="""
        References to the `OrbitalsState` sections that contain the orbitals information which are
        relevant for the `TB` model.

        Example: hydrogenated graphene with 3 atoms in the unit cell. The full list of `AtomsState` would
        be
            [
                AtomsState(chemical_symbol='C', orbitals_state=[OrbitalsState('s'), OrbitalsState('px'), OrbitalsState('py'), OrbitalsState('pz')]),
                AtomsState(chemical_symbol='C', orbitals_state=[OrbitalsState('s'), OrbitalsState('px'), OrbitalsState('py'), OrbitalsState('pz')]),
                AtomsState(chemical_symbol='H', orbitals_state=[OrbitalsState('s')]),
            ]

        The relevant orbitals for the TB model are the `'pz'` ones for each `'C'` atom. Then, we define:

            orbitals_ref= [OrbitalState('pz'), OrbitalsState('pz')]

        The relevant atoms information can be accessed from the parent AtomsState sections:
            atom_state = orbitals_ref[i].m_parent
            index = orbitals_ref[i].m_parent_index
            atom_position = orbitals_ref[i].m_parent.m_parent.positions[index]
        """,
    )

    def resolve_type(self) -> Optional[str]:
        """
        Resolves the `type` of the `TB` section if it is not already defined, and from the
        `m_def.name` of the section.

        Returns:
            (Optional[str]): The resolved `type` of the `TB` section.
        """
        return (
            self.m_def.name
            if self.m_def.name in ['DFTB', 'xTB', 'Wannier', 'SlaterKoster']
            else None
        )

    def resolve_orbital_references(
        self,
        model_systems: List[ModelSystem],
        logger: 'BoundLogger',
        model_index: int = -1,
    ) -> Optional[List[OrbitalsState]]:
        """
        Resolves the references to the `OrbitalsState` sections from the child `ModelSystem` section.

        Args:
            model_systems (List[ModelSystem]): The list of `ModelSystem` sections.
            logger (BoundLogger): The logger to log messages.
            model_index (int, optional): The `ModelSystem` section index from which resolve the references. Defaults to -1.

        Returns:
            Optional[List[OrbitalsState]]: The resolved references to the `OrbitalsState` sections.
        """
        try:
            model_system = model_systems[model_index]
        except IndexError:
            logger.warning(
                f'The `ModelSystem` section with index {model_index} was not found.'
            )
            return None

        # If `ModelSystem` is not representative, the normalization will not run
        if is_not_representative(model_system=model_system, logger=logger):
            return None

        # If `AtomicCell` is not found, the normalization will not run
        if not model_system.cell:
            logger.warning('`AtomicCell` section was not found.')
            return None
        atomic_cell = model_system.cell[0]

        # If there is no child `ModelSystem`, the normalization will not run
        atoms_state = atomic_cell.atoms_state
        model_system_child = model_system.model_system
        if not atoms_state or not model_system_child:
            logger.warning('No `AtomsState` or child `ModelSystem` section were found.')
            return None

        # We flatten the `OrbitalsState` sections from the `ModelSystem` section
        orbitals_ref = []
        for active_atom in model_system_child:
            # If the child is not an "active_atom", the normalization will not run
            if active_atom.type != 'active_atom':
                continue
            indices = active_atom.atom_indices
            for index in indices:
                try:
                    active_atoms_state = atoms_state[index]
                except IndexError:
                    logger.warning(
                        f'The `AtomsState` section with index {index} was not found.'
                    )
                    continue
                orbitals_state = active_atoms_state.orbitals_state
                for orbital in orbitals_state:
                    orbitals_ref.append(orbital)
        return orbitals_ref

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)

        # Set `name` to "TB"
        self.name = 'TB'

        # Resolve `type` to be defined by the lower level class (Wannier, DFTB, xTB or SlaterKoster) if it is not already defined
        self.type = self.resolve_type()

        # Resolve `orbitals_ref` from the info in the child `ModelSystem` section and the `OrbitalsState` sections
        model_systems = self.m_xpath('m_parent.model_system', dict=False)
        if model_systems is None:
            logger.warning(
                'Could not find the `ModelSystem` sections. References to `OrbitalsState` will not be resolved.'
            )
            return
        # This normalization only considers the last `ModelSystem` (default `model_index` argument set to -1)
        orbitals_ref = self.resolve_orbital_references(
            model_systems=model_systems, logger=logger
        )
        if orbitals_ref is not None and len(orbitals_ref) > 0 and not self.orbitals_ref:
            self.n_orbitals = len(orbitals_ref)
            self.orbitals_ref = orbitals_ref


class Wannier(TB):
    """
    A base section used to define the parameters used in a Wannier tight-binding fitting.
    """

    is_maximally_localized = Quantity(
        type=bool,
        description="""
        If the projected orbitals are maximally localized or just a single-shot projection.
        """,
    )

    localization_type = Quantity(
        type=MEnum('single_shot', 'maximally_localized'),
        description="""
        Localization type of the Wannier orbitals.
        """,
    )

    n_bloch_bands = Quantity(
        type=np.int32,
        description="""
        Number of input Bloch bands to calculate the projection matrix.
        """,
    )

    energy_window_outer = Quantity(
        type=np.float64,
        unit='electron_volt',
        shape=[2],
        description="""
        Bottom and top of the outer energy window used for the projection.
        """,
    )

    energy_window_inner = Quantity(
        type=np.float64,
        unit='electron_volt',
        shape=[2],
        description="""
        Bottom and top of the inner energy window used for the projection.
        """,
    )

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)

        # Resolve `localization_type` from `is_maximally_localized`
        if self.localization_type is None:
            if self.is_maximally_localized is not None:
                if self.is_maximally_localized:
                    self.localization_type = 'maximally_localized'
                else:
                    self.localization_type = 'single_shot'


class SlaterKosterBond(ArchiveSection):
    """
    A base section used to define the Slater-Koster bond information betwee two orbitals.
    """

    orbital_1 = Quantity(
        type=OrbitalsState,
        description="""
        Reference to the first `OrbitalsState` section.
        """,
    )

    orbital_2 = Quantity(
        type=OrbitalsState,
        description="""
        Reference to the second `OrbitalsState` section.
        """,
    )

    # ? is this the best naming
    bravais_vector = Quantity(
        type=np.int32,
        default=[0, 0, 0],
        shape=[3],
        description="""
        The Bravais vector of the cell in 3 dimensional. This is defined as the vector that connects the
        two atoms that define the Slater-Koster bond. A bond can be defined between orbitals in the
        same unit cell (bravais_vector = [0, 0, 0]) or in neighboring cells (bravais_vector = [m, n, p] with m, n, p are integers).
        Default is [0, 0, 0].
        """,
    )

    # TODO add more names and in the table
    name = Quantity(
        type=MEnum('sss', 'sps', 'sds'),
        description="""
        The name of the Slater-Koster bond. The name is composed by the `l_quantum_symbol` of the orbitals
        and the cell index. Table of possible values:

        | Value   | `orbital_1.l_quantum_symbol` | `orbital_2.l_quantum_symbol` | `bravais_vector` |
        | ------- | ---------------------------- | ---------------------------- | ------------ |
        | `'sss'` | 's' | 's' | [0, 0, 0] |
        | `'sps'` | 's' | 'p' | [0, 0, 0] |
        | `'sds'` | 's' | 'd' | [0, 0, 0] |
        """,
    )

    # ? units
    integral_value = Quantity(
        type=np.float64,
        description="""
        The Slater-Koster bond integral value.
        """,
    )

    def __init__(self, m_def: 'Section' = None, m_context: 'Context' = None, **kwargs):
        super().__init__(m_def, m_context, **kwargs)
        # TODO extend this to cover all bond names
        self._bond_name_map = {
            'sss': ['s', 's', (0, 0, 0)],
            'sps': ['s', 'p', (0, 0, 0)],
            'sds': ['s', 'd', (0, 0, 0)],
        }

    def resolve_bond_name_from_references(
        self,
        orbital_1: Optional[OrbitalsState],
        orbital_2: Optional[OrbitalsState],
        bravais_vector: Optional[tuple],
        logger: 'BoundLogger',
    ) -> Optional[str]:
        """
        Resolves the `name` of the `SlaterKosterBond` from the references to the `OrbitalsState` sections.

        Args:
            orbital_1 (Optional[OrbitalsState]): The first `OrbitalsState` section.
            orbital_2 (Optional[OrbitalsState]): The second `OrbitalsState` section.
            bravais_vector (Optional[tuple]): The bravais vector of the cell.
            logger (BoundLogger): The logger to log messages.

        Returns:
            (Optional[str]): The resolved `name` of the `SlaterKosterBond`.
        """
        # Initial check
        if orbital_1 is None or orbital_2 is None:
            logger.warning('The `OrbitalsState` sections are not defined.')
            return None
        if bravais_vector is None:
            logger.warning('The `bravais_vector` is not defined.')
            return None

        # Check for `l_quantum_symbol` in `OrbitalsState` sections
        if orbital_1.l_quantum_symbol is None or orbital_2.l_quantum_symbol is None:
            logger.warning(
                'The `l_quantum_symbol` of the `OrbitalsState` bonds are not defined.'
            )
            return None

        bond_name = None
        value = [orbital_1.l_quantum_symbol, orbital_2.l_quantum_symbol, bravais_vector]
        # Check if `value` is found in the `self._bond_name_map` and return the key
        for key, val in self._bond_name_map.items():
            if val == value:
                bond_name = key
                break
        return bond_name

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)

        # Resolve the SK bond `name` from the `OrbitalsState` references and the `bravais_vector`
        if self.orbital_1 and self.orbital_2 and self.bravais_vector is not None:
            if self.bravais_vector is not None:
                bravais_vector = tuple(self.bravais_vector)  # transformed for comparing
            self.name = self.resolve_bond_name_from_references(
                orbital_1=self.orbital_1,
                orbital_2=self.orbital_2,
                bravais_vector=bravais_vector,
                logger=logger,
            )


class SlaterKoster(TB):
    """
    A base section used to define the parameters used in a Slater-Koster tight-binding fitting.
    """

    bonds = SubSection(sub_section=SlaterKosterBond.m_def, repeats=True)

    overlaps = SubSection(sub_section=SlaterKosterBond.m_def, repeats=True)

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


class xTB(TB):
    """
    A base section used to define the parameters used in an extended tight-binding (xTB) calculation.
    """

    # ? Deprecate this

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


class Photon(ArchiveSection):
    """
    A base section used to define parameters of a photon, typically used for optical responses.
    """

    # TODO check other options and add specific refs
    multipole_type = Quantity(
        type=MEnum('dipolar', 'quadrupolar', 'NRIXS', 'Raman'),
        description="""
        Type used for the multipolar expansion: dipole, quadrupole, NRIXS, Raman, etc.
        """,
    )

    polarization = Quantity(
        type=np.float64,
        shape=[3],
        description="""
        Direction of the photon polarization in cartesian coordinates.
        """,
    )

    energy = Quantity(
        type=np.float64,
        unit='joule',
        description="""
        Photon energy.
        """,
    )

    momentum_transfer = Quantity(
        type=np.float64,
        shape=[3],
        description="""
        Momentum transfer to the lattice. This quanitity is important for inelastic scatterings, like
        the ones happening in quadrupolar, Raman, or NRIXS processes.
        """,
    )

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)

        # Add warning in case `multipole_type` and `momentum_transfer` are not consistent
        if (
            self.multipole_type in ['quadrupolar', 'NRIXS', 'Raman']
            and self.momentum_transfer is None
        ):
            logger.warning(
                'The `Photon.momentum_transfer` is not defined but the `Photon.multipole_type` describes inelastic scattering processes.'
            )


class ExcitedStateMethodology(ModelMethodElectronic):
    """
    A base section used to define the parameters typical of excited-state calculations. "ExcitedStateMethodology"
    mainly refers to methodologies which consider many-body effects as a perturbation of the original
    DFT Hamiltonian. These are: GW, TDDFT, BSE.
    """

    type = Quantity(
        type=str,
        description="""
        Identifier used to further specify the type of model Hamiltonian.
        """,
        a_eln=ELNAnnotation(component='StringEditQuantity'),
    )

    n_states = Quantity(
        type=np.int32,
        description="""
        Number of states used to calculate the excitations.
        """,
        a_eln=ELNAnnotation(component='NumberEditQuantity'),
    )

    n_empty_states = Quantity(
        type=np.int32,
        description="""
        Number of empty states used to calculate the excitations. This quantity is complementary to `n_states`.
        """,
        a_eln=ELNAnnotation(component='NumberEditQuantity'),
    )

    broadening = Quantity(
        type=np.float64,
        unit='joule',
        description="""
        Lifetime broadening applied to the spectra in full-width at half maximum for excited-state calculations.
        """,
        a_eln=ELNAnnotation(component='NumberEditQuantity'),
    )

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


class Screening(ExcitedStateMethodology):
    """
    A base section used to define the parameters that define the calculation of screening. This is usually done in
    RPA and linear response.
    """

    dielectric_infinity = Quantity(
        type=np.int32,
        description="""
        Value of the static dielectric constant at infinite q. For metals, this is infinite
        (or a very large value), while for insulators is finite.
        """,
        a_eln=ELNAnnotation(component='NumberEditQuantity'),
    )

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


class GW(ExcitedStateMethodology):
    """
    A base section used to define the parameters of a GW calculation.
    """

    type = Quantity(
        type=MEnum(
            'G0W0',
            'scGW',
            'scGW0',
            'scG0W',
            'ev-scGW0',
            'ev-scGW',
            'qp-scGW0',
            'qp-scGW',
        ),
        description="""
        GW Hedin's self-consistency cycle:

        | Name      | Description                      | Reference             |
        | --------- | -------------------------------- | --------------------- |
        | `'G0W0'`  | single-shot                      | https://journals.aps.org/prb/abstract/10.1103/PhysRevB.74.035101 |
        | `'scGW'`  | self-consistent G and W               | https://journals.aps.org/prb/abstract/10.1103/PhysRevB.75.235102 |
        | `'scGW0'` | self-consistent G with fixed W0  | https://journals.aps.org/prb/abstract/10.1103/PhysRevB.54.8411 |
        | `'scG0W'` | self-consistent W with fixed G0  | -                     |
        | `'ev-scGW0'`  | eigenvalues self-consistent G with fixed W0   | https://journals.aps.org/prb/abstract/10.1103/PhysRevB.34.5390 |
        | `'ev-scGW'`  | eigenvalues self-consistent G and W   | https://journals.aps.org/prb/abstract/10.1103/PhysRevB.74.045102 |
        | `'qp-scGW0'`  | quasiparticle self-consistent G with fixed W0 | https://journals.aps.org/prb/abstract/10.1103/PhysRevB.76.115109 |
        | `'qp-scGW'`  | quasiparticle self-consistent G and W | https://journals.aps.org/prl/abstract/10.1103/PhysRevLett.96.226402 |
        """,
        a_eln=ELNAnnotation(component='EnumEditQuantity'),
    )

    analytical_continuation = Quantity(
        type=MEnum(
            'pade',
            'contour_deformation',
            'ppm_GodbyNeeds',
            'ppm_HybertsenLouie',
            'ppm_vonderLindenHorsh',
            'ppm_FaridEngel',
            'multi_pole',
        ),
        description="""
        Analytical continuation approximations of the GW self-energy:

        | Name           | Description         | Reference                        |
        | -------------- | ------------------- | -------------------------------- |
        | `'pade'` | Pade's approximant  | https://link.springer.com/article/10.1007/BF00655090 |
        | `'contour_deformation'` | Contour deformation | https://journals.aps.org/prb/abstract/10.1103/PhysRevB.67.155208 |
        | `'ppm_GodbyNeeds'` | Godby-Needs plasmon-pole model | https://journals.aps.org/prl/abstract/10.1103/PhysRevLett.62.1169 |
        | `'ppm_HybertsenLouie'` | Hybertsen and Louie plasmon-pole model | https://journals.aps.org/prb/abstract/10.1103/PhysRevB.34.5390 |
        | `'ppm_vonderLindenHorsh'` | von der Linden and P. Horsh plasmon-pole model | https://journals.aps.org/prb/abstract/10.1103/PhysRevB.37.8351 |
        | `'ppm_FaridEngel'` | Farid and Engel plasmon-pole model  | https://journals.aps.org/prb/abstract/10.1103/PhysRevB.47.15931 |
        | `'multi_pole'` | Multi-pole fitting  | https://journals.aps.org/prl/abstract/10.1103/PhysRevLett.74.1827 |
        """,
        a_eln=ELNAnnotation(component='EnumEditQuantity'),
    )

    # TODO improve description
    interval_qp_corrections = Quantity(
        type=np.int32,
        shape=[2],
        description="""
        Band indices (in an interval) for which the GW quasiparticle corrections are calculated.
        """,
    )

    screening_ref = Quantity(
        type=Screening,
        description="""
        Reference to the `Screening` section that the GW calculation used to obtain the screened Coulomb interactions.
        """,
        a_eln=ELNAnnotation(component='ReferenceEditQuantity'),
    )

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


class BSE(ExcitedStateMethodology):
    """
    A base section used to define the parameters of a BSE calculation.
    """

    # ? does RPA relates with `screening_ref`?
    type = Quantity(
        type=MEnum('Singlet', 'Triplet', 'IP', 'RPA'),
        description="""
        Type of BSE hamiltonian solved:

            H_BSE = H_diagonal + 2 * gx * Hx - gc * Hc

        where gx, gc specifies the type.

        Online resources for the theory:
        - http://exciting.wikidot.com/carbon-excited-states-from-bse#toc1
        - https://www.vasp.at/wiki/index.php/Bethe-Salpeter-equations_calculations
        - https://docs.abinit.org/theory/bse/
        - https://www.yambo-code.eu/wiki/index.php/Bethe-Salpeter_kernel

        | Name | Description |
        | --------- | ----------------------- |
        | `'Singlet'` | gx = 1, gc = 1 |
        | `'Triplet'` | gx = 0, gc = 1 |
        | `'IP'` | Independent-particle approach |
        | `'RPA'` | Random Phase Approximation |
        """,
        a_eln=ELNAnnotation(component='EnumEditQuantity'),
    )

    solver = Quantity(
        type=MEnum('Full-diagonalization', 'Lanczos-Haydock', 'GMRES', 'SLEPc', 'TDA'),
        description="""
        Solver algotithm used to diagonalize the BSE Hamiltonian.

        | Name | Description | Reference |
        | --------- | ----------------------- | ----------- |
        | `'Full-diagonalization'` | Full diagonalization of the BSE Hamiltonian | - |
        | `'Lanczos-Haydock'` | Subspace iterative Lanczos-Haydock algorithm | https://doi.org/10.1103/PhysRevB.59.5441 |
        | `'GMRES'` | Generalized minimal residual method | https://doi.org/10.1137/0907058 |
        | `'SLEPc'` | Scalable Library for Eigenvalue Problem Computations | https://slepc.upv.es/ |
        | `'TDA'` | Tamm-Dancoff approximation | https://doi.org/10.1016/S0009-2614(99)01149-5 |
        """,
        a_eln=ELNAnnotation(component='EnumEditQuantity'),
    )

    screening_ref = Quantity(
        type=Screening,
        description="""
        Reference to the `Screening` section that the BSE calculation used to obtain the screened Coulomb interactions.
        """,
        a_eln=ELNAnnotation(component='ReferenceEditQuantity'),
    )

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


# ? Is this class really necessary or should go in outputs.py?
class CoreHoleSpectra(ModelMethodElectronic):
    """
    A base section used to define the parameters used in a core-hole spectra calculation. This
    also contains reference to the specific methodological section (DFT, BSE) used to obtain the core-hole spectra.
    """

    # # TODO add examples
    # solver = Quantity(
    #     type=str,
    #     description="""
    #     Solver algorithm used for the core-hole spectra.
    #     """,
    #     a_eln=ELNAnnotation(component="StringEditQuantity"),
    # )

    type = Quantity(
        type=MEnum('absorption', 'emission'),
        description="""
        Type of the CoreHole excitation spectra calculated, either "absorption" or "emission".
        """,
        a_eln=ELNAnnotation(component='EnumEditQuantity'),
    )

    edge = Quantity(
        type=MEnum(
            'K',
            'L1',
            'L2',
            'L3',
            'L23',
            'M1',
            'M2',
            'M3',
            'M23',
            'M4',
            'M5',
            'M45',
            'N1',
            'N2',
            'N3',
            'N23',
            'N4',
            'N5',
            'N45',
        ),
        description="""
        Edge label of the excited core-hole. This is obtained by normalization by using `core_hole_ref`.
        """,
    )

    core_hole_ref = Quantity(
        type=CoreHole,
        description="""
        Reference to the `CoreHole` section that contains the information of the edge of the excited core-hole.
        """,
        a_eln=ELNAnnotation(component='ReferenceEditQuantity'),
    )

    excited_state_method_ref = Quantity(
        type=ModelMethodElectronic,
        description="""
        Reference to the `ModelMethodElectronic` section (e.g., `DFT` or `BSE`) that was used to obtain the core-hole spectra.
        """,
        a_eln=ELNAnnotation(component='ReferenceEditQuantity'),
    )

    # TODO add normalization to obtain `edge`

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


class DMFT(ModelMethodElectronic):
    """
    A base section used to define the parameters of a DMFT calculation.
    """

    impurity_solver = Quantity(
        type=MEnum(
            'CT-INT',
            'CT-HYB',
            'CT-AUX',
            'ED',
            'NRG',
            'MPS',
            'IPT',
            'NCA',
            'OCA',
            'slave_bosons',
            'hubbard_I',
        ),
        description="""
        Impurity solver method used in the DMFT loop:

        | Name              | Reference                            |
        | ----------------- | ------------------------------------ |
        | `'CT-INT'`        | https://link.springer.com/article/10.1134/1.1800216 |
        | `'CT-HYB'`        | https://journals.aps.org/prl/abstract/10.1103/PhysRevLett.97.076405 |
        | `'CT-AUX'`        | https://iopscience.iop.org/article/10.1209/0295-5075/82/57003 |
        | `'ED'`            | https://journals.aps.org/prl/abstract/10.1103/PhysRevLett.72.1545 |
        | `'NRG'`           | https://journals.aps.org/rmp/abstract/10.1103/RevModPhys.80.395 |
        | `'MPS'`           | https://journals.aps.org/prb/abstract/10.1103/PhysRevB.90.045144 |
        | `'IPT'`           | https://journals.aps.org/prb/abstract/10.1103/PhysRevB.45.6479 |
        | `'NCA'`           | https://journals.aps.org/prb/abstract/10.1103/PhysRevB.47.3553 |
        | `'OCA'`           | https://journals.aps.org/prb/abstract/10.1103/PhysRevB.47.3553 |
        | `'slave_bosons'`  | https://journals.aps.org/prl/abstract/10.1103/PhysRevLett.57.1362 |
        | `'hubbard_I'`     | https://iopscience.iop.org/article/10.1088/0953-8984/24/7/075604 |
        """,
    )

    n_impurities = Quantity(
        type=np.int32,
        description="""
        Number of impurities mapped from the correlated atoms in the unit cell. This defines whether
        the DMFT calculation is done in a single-impurity or multi-impurity run.
        """,
    )

    n_orbitals = Quantity(
        type=np.int32,
        shape=['n_impurities'],
        description="""
        Number of correlated orbitals per impurity.
        """,
    )

    orbitals_ref = Quantity(
        type=OrbitalsState,
        shape=['n_orbitals'],
        description="""
        References to the `OrbitalsState` sections that contain the orbitals information which are
        relevant for the `DMFT` calculation.

        Example: hydrogenated graphene with 3 atoms in the unit cell. The full list of `AtomsState` would
        be
            [
                AtomsState(chemical_symbol='C', orbitals_state=[OrbitalsState('s'), OrbitalsState('px'), OrbitalsState('py'), OrbitalsState('pz')]),
                AtomsState(chemical_symbol='C', orbitals_state=[OrbitalsState('s'), OrbitalsState('px'), OrbitalsState('py'), OrbitalsState('pz')]),
                AtomsState(chemical_symbol='H', orbitals_state=[OrbitalsState('s')]),
            ]

        The relevant orbitals for the TB model are the `'pz'` ones for each `'C'` atom. Then, we define:

            orbitals_ref= [OrbitalState('pz'), OrbitalsState('pz')]

        The relevant impurities information can be accesed from the parent AtomsState sections:
            impurity_state = orbitals_ref[i].m_parent
            index = orbitals_ref[i].m_parent_index
            impurity_position = orbitals_ref[i].m_parent.m_parent.positions[index]
        """,
    )

    # ? Improve this with `orbitals_ref.occupation` and possibly a function?
    n_electrons = Quantity(
        type=np.float64,
        shape=['n_impurities'],
        description="""
        Initial number of valence electrons per impurity.
        """,
    )

    inverse_temperature = Quantity(
        type=np.float64,
        unit='1/joule',
        description="""
        Inverse temperature = 1/(kB*T).
        """,
    )

    # ? Check this once magnetic states are better covered in the schema. This will be probably under `ModelSystem`
    # ? by checking the spins in `AtomsState` for the `AtomicCell`
    # ! Check solid_dmft example by using magmom (atomic magnetic moments), and improve on AtomsState to include such moments
    magnetic_state = Quantity(
        type=MEnum('paramagnetic', 'ferromagnetic', 'antiferromagnetic'),
        description="""
        Magnetic state in which the DMFT calculation is done. This quantity can be obtained from
        `orbitals_ref` and their spin state.
        """,
    )

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from nomad_simulations.schema_packages.properties import FermiLevel, ChemicalPotential


class TestFermiLevel:
    """
    Test the `FermiLevel` class defined in `properties/energies.py`.
    """

    # ! Include this initial `test_default_quantities` method when testing your PhysicalProperty classes
    def test_default_quantities(self):
        """
        Test the default quantities assigned when creating an instance of the `FermiLevel` class.
        """
        fermi_level = FermiLevel()
        assert fermi_level.iri == 'http://fairmat-nfdi.eu/taxonomy/FermiLevel'
        assert fermi_level.name == 'FermiLevel'
        assert fermi_level.rank == []


class TestChemicalPotential:
    """
    Test the `ChemicalPotential` class defined in `properties/energies.py`.
    """

    # ! Include this initial `test_default_quantities` method when testing your PhysicalProperty classes
    def test_default_quantities(self):
        """
        Test the default quantities assigned when creating an instance of the `ChemicalPotential` class.
        """
        chemical_potential = ChemicalPotential()
        assert (
            chemical_potential.iri
            == 'http://fairmat-nfdi.eu/taxonomy/ChemicalPotential'
        )
        assert chemical_potential.name == 'ChemicalPotential'
        assert chemical_potential.rank == []



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import TYPE_CHECKING

import numpy as np

from nomad.metainfo import Quantity

if TYPE_CHECKING:
    from nomad.metainfo import Section, Context
    from nomad.datamodel.datamodel import EntryArchive
    from structlog.stdlib import BoundLogger

from nomad_simulations.schema_packages.physical_property import PhysicalProperty


class HoppingMatrix(PhysicalProperty):
    """
    Transition probability between two atomic orbitals in a tight-binding model.
    """

    iri = 'http://fairmat-nfdi.eu/taxonomy/HoppingMatrix'

    n_orbitals = Quantity(
        type=np.int32,
        description="""
        Number of orbitals in the tight-binding model.
        """,
    )

    degeneracy_factors = Quantity(
        type=np.int32,
        shape=['*'],
        description="""
        Degeneracy of each Wigner-Seitz point.
        """,
    )

    value = Quantity(
        type=np.complex128,
        unit='joule',
        description="""
        Value of the hopping matrix in joules. The elements are complex numbers defined for each Wigner-Seitz point and
        each pair of orbitals; thus, `rank = [n_orbitals, n_orbitals]`. Note this contains also the onsite values, i.e.,
        it includes the Wigner-Seitz point (0, 0, 0), hence the `CrystalFieldSplitting` values.
        """,
    )

    def __init__(
        self, m_def: 'Section' = None, m_context: 'Context' = None, **kwargs
    ) -> None:
        super().__init__(m_def, m_context, **kwargs)
        # ! n_orbitals need to be set up during initialization of the class
        self.rank = [int(kwargs.get('n_orbitals')), int(kwargs.get('n_orbitals'))]
        self.name = self.m_def.name

    # TODO add normalization to extract DOS, band structure, etc, properties from `HoppingMatrix`

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


class CrystalFieldSplitting(PhysicalProperty):
    """
    Energy difference between the degenerated orbitals of an ion in a crystal field environment.
    """

    iri = 'http://fairmat-nfdi.eu/taxonomy/CrystalFieldSplitting'

    n_orbitals = Quantity(
        type=np.int32,
        description="""
        Number of orbitals in the tight-binding model.
        """,
    )

    value = Quantity(
        type=np.float64,
        unit='joule',
        description="""
        Value of the crystal field splittings in joules. This is the intra-orbital local contribution, i.e., the same orbital
        at the same Wigner-Seitz point (0, 0, 0).
        """,
    )

    def __init__(
        self, m_def: 'Section' = None, m_context: 'Context' = None, **kwargs
    ) -> None:
        super().__init__(m_def, m_context, **kwargs)
        # ! `n_orbitals` need to be set up during initialization of the class
        self.rank = [int(kwargs.get('n_orbitals'))]
        self.name = self.m_def.name

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import TYPE_CHECKING, Any, Dict, Optional, Union

import ase
import numpy as np
import pint

from nomad.datamodel.data import ArchiveSection
from nomad.datamodel.metainfo.annotations import ELNAnnotation
from nomad.datamodel.metainfo.basesections import Entity
from nomad.metainfo import MEnum, Quantity, SubSection
from nomad.units import ureg

if TYPE_CHECKING:
    from nomad.metainfo import Section, Context
    from nomad.datamodel.datamodel import EntryArchive
    from structlog.stdlib import BoundLogger

from nomad_simulations.schema_packages.utils import RussellSaundersState


class OrbitalsState(Entity):
    """
    A base section used to define the orbital state of an atom.
    """

    # TODO add check for `j_quantum_number` and `mj_quantum_number`
    # TODO: add the relativistic kappa_quantum_number

    n_quantum_number = Quantity(
        type=np.int32,
        description="""
        Principal quantum number of the orbital state.
        """,
    )

    l_quantum_number = Quantity(
        type=np.int32,
        description="""
        Azimuthal quantum number of the orbital state. This quantity is equivalent to `l_quantum_symbol`.
        """,
    )

    l_quantum_symbol = Quantity(
        type=str,
        description="""
        Azimuthal quantum symbol of the orbital state, "s", "p", "d", "f", etc. This
        quantity is equivalent to `l_quantum_number`.
        """,
    )

    ml_quantum_number = Quantity(
        type=np.int32,
        description="""
        Azimuthal projection number of the `l` vector. This quantity is equivalent to `ml_quantum_symbol`.
        """,
    )

    ml_quantum_symbol = Quantity(
        type=str,
        description="""
        Azimuthal projection symbol of the `l` vector, "x", "y", "z", etc. This quantity is equivalent
        to `ml_quantum_number`.
        """,
    )

    j_quantum_number = Quantity(
        type=np.float64,
        shape=['1..2'],
        description="""
        Total angular momentum quantum number $j = |l-s| ... l+s$. Necessary with strong
        L-S coupling or non-collinear spin systems.
        """,
    )

    mj_quantum_number = Quantity(
        type=np.float64,
        shape=['*'],
        description="""
        Azimuthal projection of the `j` vector. Necessary with strong L-S coupling or
        non-collinear spin systems.
        """,
    )

    ms_quantum_number = Quantity(
        type=np.float64,
        description="""
        Spin quantum number. Set to -1 for spin down and +1 for spin up. In non-collinear spin
        systems, the projection axis $z$ should also be defined.
        """,
    )

    ms_quantum_symbol = Quantity(
        type=MEnum('down', 'up'),
        description="""
        Spin quantum symbol. Set to 'down' for spin down and 'up' for spin up. In non-collinear
        spin systems, the projection axis $z$ should also be defined.
        """,
    )

    degeneracy = Quantity(
        type=np.int32,
        description="""
        The degeneracy of the orbital state. The degeneracy is the number of states with the same
        energy. It is equal to 2 * l + 1 for non-relativistic systems and 2 * j + 1 for
        relativistic systems, if ms_quantum_number is defined (otherwise a factor of 2 is included).
        """,
    )

    occupation = Quantity(
        type=np.float64,
        description="""
        The number of electrons in the orbital state. The state is fully occupied if
        occupation = degeneracy.
        """,
    )

    def __init__(self, m_def: 'Section' = None, m_context: 'Context' = None, **kwargs):
        super().__init__(m_def, m_context, **kwargs)
        self._orbitals = {
            -1: dict(zip(range(4), ('s', 'p', 'd', 'f'))),
            0: {0: ''},
            1: dict(zip(range(-1, 2), ('x', 'z', 'y'))),
            2: dict(zip(range(-2, 3), ('xy', 'xz', 'z^2', 'yz', 'x^2-y^2'))),
            3: dict(
                zip(
                    range(-3, 4),
                    (
                        'x(x^2-3y^2)',
                        'xyz',
                        'xz^2',
                        'z^3',
                        'yz^2',
                        'z(x^2-y^2)',
                        'y(3x^2-y^2)',
                    ),
                )
            ),
        }
        self._orbitals_map: Dict[str, Any] = {
            'l_symbols': self._orbitals[-1],
            'ml_symbols': {i: self._orbitals[i] for i in range(4)},
            'ms_symbols': dict(zip((-0.5, 0.5), ('down', 'up'))),
            'l_numbers': {v: k for k, v in self._orbitals[-1].items()},
            'ml_numbers': {
                k: {v: k for k, v in self._orbitals[k].items()} for k in range(4)
            },
            'ms_numbers': dict(zip(('down', 'up'), (-0.5, 0.5))),
        }

    def validate_quantum_numbers(self, logger: 'BoundLogger') -> bool:
        """
        Validate the quantum numbers (`n`, `l`, `ml`, `ms`) by checking if they are physically sensible.

        Args:
            logger (BoundLogger): The logger to log messages.

        Returns:
            (bool): True if the quantum numbers are physically sensible, False otherwise.
        """
        if self.n_quantum_number is not None and self.n_quantum_number < 1:
            logger.error('The `n_quantum_number` must be greater than 0.')
            return False
        if self.l_quantum_number is not None and self.l_quantum_number < 1:
            logger.error('The `l_quantum_number` must be greater than 0.')
            return False
        if self.ml_quantum_number is not None and (
            self.ml_quantum_number < -self.l_quantum_number
            or self.ml_quantum_number > self.l_quantum_number
        ):
            logger.error(
                'The `ml_quantum_number` must be between `-l_quantum_number` and `l_quantum_number`.'
            )
            return False
        if self.ms_quantum_number is not None and self.ms_quantum_number not in [
            -0.5,
            0.5,
        ]:
            logger.error('The `ms_quantum_number` must be -0.5 or 0.5.')
            return False
        return True

    def resolve_number_and_symbol(
        self, quantum_name: str, quantum_type: str, logger: 'BoundLogger'
    ) -> Optional[Union[str, int]]:
        """
        Resolves the quantum number or symbol from the `self._orbitals_map` on the passed `quantum_type`.
        `quantum_type` can be either 'number' or 'symbol'. If the quantum type is not found, then the countertype
        (e.g., quantum_type == 'number' => countertype == 'symbol') is used to resolve it.

        Args:
            quantum_name (str): The quantum name to resolve. Can be 'l', 'ml' or 'ms'.
            quantum_type (str): The type of the quantum name. Can be 'number' or 'symbol'.
            logger (BoundLogger): The logger to log messages.

        Returns:
            (Optional[Union[str, int]]): The quantum number or symbol resolved from the orbitals_map.
        """
        if quantum_name not in ['l', 'ml', 'ms']:
            logger.warning("The quantum_name is not recognized. Try 'l', 'ml' or 'ms'.")
            return None
        if quantum_type not in ['number', 'symbol']:
            logger.warning(
                f"The quantum_type {quantum_type} is not recognized. Try 'number' or 'symbol'."
            )
            return None

        # Check if quantity already exists
        quantity = getattr(self, f'{quantum_name}_quantum_{quantum_type}')
        if quantity is not None:
            return quantity

        # If not, check whether the countertype exists
        _countertype_map = {
            'number': 'symbol',
            'symbol': 'number',
        }
        other_quantity = getattr(
            self, f'{quantum_name}_quantum_{_countertype_map[quantum_type]}'
        )
        if other_quantity is None:
            return None

        # If the counterpart exists, then resolve the quantity from the orbitals_map
        orbital_quantity = self._orbitals_map.get(f'{quantum_name}_{quantum_type}s', {})
        if quantum_name in ['l', 'ms']:
            quantity = orbital_quantity.get(other_quantity)
        elif quantum_name == 'ml':
            if self.l_quantum_number is None:
                return None
            quantity = orbital_quantity.get(self.l_quantum_number, {}).get(
                other_quantity
            )
        return quantity

    def resolve_degeneracy(self) -> Optional[int]:
        """
        Resolves the degeneracy of the orbital state. If `j_quantum_number` is not defined, then the
        degeneracy is computed from the `l_quantum_number` and `ml_quantum_number`. If `j_quantum_number`
        is defined, then the degeneracy is computed from the `j_quantum_number` and `mj_quantum_number`.
        There is a factor of 2 included if `ms_quantum_number` is not defined (for orbitals which
        are spin-degenerated).

        Returns:
            (Optional[int]): The degeneracy of the orbital state.
        """
        degeneracy = None
        if (
            self.l_quantum_number
            and self.ml_quantum_number is None
            and self.j_quantum_number is None
        ):
            if self.ms_quantum_number:
                degeneracy = 2 * self.l_quantum_number + 1
            else:
                degeneracy = 2 * (2 * self.l_quantum_number + 1)
        elif (
            self.l_quantum_number
            and self.ml_quantum_number
            and self.j_quantum_number is None
        ):
            if self.ms_quantum_number:
                degeneracy = 1
            else:
                degeneracy = 2
        elif self.j_quantum_number is not None:
            degeneracy = 0
            for jj in self.j_quantum_number:
                if self.mj_quantum_number is not None:
                    mjs = RussellSaundersState.generate_MJs(
                        J=self.j_quantum_number[0], rising=True
                    )
                    degeneracy += len(
                        [mj for mj in mjs if mj in self.mj_quantum_number]
                    )
                else:
                    degeneracy += RussellSaundersState(J=jj, occ=1).degeneracy
        return degeneracy

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)

        # General checks for physical quantum numbers and symbols
        if not self.validate_quantum_numbers(logger=logger):
            logger.error('The quantum numbers are not physical.')
            return

        # Resolving the quantum numbers and symbols if not available
        for quantum_name in ['l', 'ml', 'ms']:
            for quantum_type in ['number', 'symbol']:
                quantity = self.resolve_number_and_symbol(
                    quantum_name=quantum_name, quantum_type=quantum_type, logger=logger
                )
                if getattr(self, f'{quantum_name}_quantum_{quantum_type}') is None:
                    setattr(self, f'{quantum_name}_quantum_{quantum_type}', quantity)

        # Resolve the degeneracy
        if self.degeneracy is None:
            self.degeneracy = self.resolve_degeneracy()


class CoreHole(ArchiveSection):
    """
    A base section used to define the core-hole state of an atom by referencing the `OrbitalsState`
    section where the core-hole was generated.
    """

    orbital_ref = Quantity(
        type=OrbitalsState,
        description="""
        Reference to the OrbitalsState section that is used as a basis to obtain the `CoreHole` section.
        """,
        a_eln=ELNAnnotation(component='ReferenceEditQuantity'),
    )

    n_excited_electrons = Quantity(
        type=np.float64,
        description="""
        The electron charge excited for modelling purposes. This is a number between 0 and 1 (Janak state).
        If `dscf_state` is set to 'initial', then this quantity is set to None (but assumed to be excited
        to an excited state).
        """,
    )

    dscf_state = Quantity(
        type=MEnum('initial', 'final'),
        description="""
        Tag used to identify the role in the workflow of the same name. Allowed values are 'initial'
        (not to be confused with the _initial-state approximation_) and 'final'. If 'initial'
        is used, then `n_excited_electrons` is set to None and the `orbital_ref.degeneracy` is
        set to 1.
        """,
    )

    def resolve_occupation(self, logger: 'BoundLogger') -> Optional[np.float64]:
        """
        Resolves the occupation of the orbital state. The occupation is resolved from the degeneracy
        and the number of excited electrons.

        Args:
            logger (BoundLogger): The logger to log messages.

        Returns:
            (Optional[np.float64]): The occupation of the active orbital state.
        """
        if self.orbital_ref is None or self.n_excited_electrons is None:
            logger.warning(
                'Cannot resolve occupation without `orbital_ref` or `n_excited_electrons`.'
            )
            return None
        if self.orbital_ref.occupation is None:
            degeneracy = self.orbital_ref.resolve_degeneracy()
            if degeneracy is None:
                logger.warning('Cannot resolve occupation without `degeneracy`.')
                return None
            return degeneracy - self.n_excited_electrons
        return None

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)

        # Check if n_excited_electrons is between 0 and 1
        if self.n_excited_electrons < 0 or self.n_excited_electrons > 1:
            logger.error('Number of excited electrons must be between 0 and 1.')
            return

        # Resolve the occupation of the active orbital state
        if self.orbital_ref is not None:
            # If dscf_state is 'initial', then n_excited_electrons is set to 0
            if self.dscf_state == 'initial':
                self.n_excited_electrons = None
                self.orbital_ref.degeneracy = 1
            if self.orbital_ref.occupation is None:
                self.orbital_ref.occupation = self.resolve_occupation(logger=logger)


class HubbardInteractions(ArchiveSection):
    """
    A base section to define the Hubbard interactions of the system.
    """

    # TODO (@JosePizarro3 note): we need to have checks for when a `ModelSystem` is spin rotational invariant (then we only need to pass `u_interaction` and `j_hunds_coupling` and resolve the other quantities)

    n_orbitals = Quantity(
        type=np.int32,
        description="""
        Number of orbitals used to define the Hubbard interactions.
        """,
    )

    orbitals_ref = Quantity(
        type=OrbitalsState,
        shape=['n_orbitals'],
        description="""
        Reference to the `OrbitalsState` sections that are used as a basis to obtain the Hubbard
        interaction matrices.
        """,
    )

    u_matrix = Quantity(
        type=np.float64,
        shape=['n_orbitals', 'n_orbitals'],
        unit='joule',
        description="""
        Value of the local Hubbard interaction matrix. The order of the rows and columns coincide
        with the elements in `orbital_ref`.
        """,
    )

    u_interaction = Quantity(
        type=np.float64,
        unit='joule',
        description="""
        Value of the (intraorbital) Hubbard interaction
        """,
        a_eln=ELNAnnotation(component='NumberEditQuantity'),
    )

    j_hunds_coupling = Quantity(
        type=np.float64,
        unit='joule',
        description="""
        Value of the (interorbital) Hund's coupling.
        """,
        a_eln=ELNAnnotation(component='NumberEditQuantity'),
    )

    u_interorbital_interaction = Quantity(
        type=np.float64,
        unit='joule',
        description="""
        Value of the (interorbital) Coulomb interaction. In rotational invariant systems,
        u_interorbital_interaction = u_interaction - 2 * j_hunds_coupling.
        """,
        a_eln=ELNAnnotation(component='NumberEditQuantity'),
    )

    j_local_exchange_interaction = Quantity(
        type=np.float64,
        unit='joule',
        description="""
        Value of the exchange interaction. In rotational invariant systems, j_local_exchange_interaction = j_hunds_coupling.
        """,
        a_eln=ELNAnnotation(component='NumberEditQuantity'),
    )

    u_effective = Quantity(
        type=np.float64,
        unit='joule',
        description="""
        Value of the effective U parameter (u_interaction - j_local_exchange_interaction).
        """,
    )

    slater_integrals = Quantity(
        type=np.float64,
        shape=[3],
        unit='joule',
        description="""
        Value of the Slater integrals [F0, F2, F4] in spherical harmonics used to derive
        the local Hubbard interactions:

            u_interaction = ((2.0 / 7.0) ** 2) * (F0 + 5.0 * F2 + 9.0 * F4) / (4.0*np.pi)

            u_interorbital_interaction = ((2.0 / 7.0) ** 2) * (F0 - 5.0 * F2 + 3.0 * 0.5 * F4) / (4.0*np.pi)

            j_hunds_coupling = ((2.0 / 7.0) ** 2) * (5.0 * F2 + 15.0 * 0.25 * F4) / (4.0*np.pi)

        See e.g., Elbio Dagotto, Nanoscale Phase Separation and Colossal Magnetoresistance,
        Chapter 4, Springer Berlin (2003).
        """,
    )

    double_counting_correction = Quantity(
        type=str,
        description="""
        Name of the double counting correction algorithm applied.
        """,
        a_eln=ELNAnnotation(component='StringEditQuantity'),
    )

    def resolve_u_interactions(self, logger: 'BoundLogger') -> Optional[tuple]:
        """
        Resolves the Hubbard interactions (u_interaction, u_interorbital_interaction, j_hunds_coupling)
        from the Slater integrals (F0, F2, F4) in the units defined for the Quantity.

        Args:
            logger (BoundLogger): The logger to log messages.

        Returns:
            (Optional[tuple]): The Hubbard interactions (u_interaction, u_interorbital_interaction, j_hunds_coupling).
        """
        if self.slater_integrals is None or len(self.slater_integrals) != 3:
            logger.warning(
                'Could not find `slater_integrals` or the length is not three.'
            )
            return None, None, None
        f0 = self.slater_integrals[0]
        f2 = self.slater_integrals[1]
        f4 = self.slater_integrals[2]
        u_interaction = ((2.0 / 7.0) ** 2) * (f0 + 5.0 * f2 + 9.0 * f4) / (4.0 * np.pi)
        u_interorbital_interaction = (
            ((2.0 / 7.0) ** 2) * (f0 - 5.0 * f2 + 3.0 * f4 / 2.0) / (4.0 * np.pi)
        )
        j_hunds_coupling = (
            ((2.0 / 7.0) ** 2) * (5.0 * f2 + 15.0 * f4 / 4.0) / (4.0 * np.pi)
        )
        return u_interaction, u_interorbital_interaction, j_hunds_coupling

    def resolve_u_effective(self, logger: 'BoundLogger') -> Optional[pint.Quantity]:
        """
        Resolves the effective U parameter (u_interaction - j_local_exchange_interaction).

        Args:
            logger (BoundLogger): The logger to log messages.

        Returns:
            (Optional[pint.Quantity]): The effective U parameter.
        """
        if self.u_interaction is None:
            logger.warning('Could not find `HubbardInteractions.u_interaction`.')
            return None
        if self.u_interaction.magnitude < 0.0:
            logger.error('The `HubbardInteractions.u_interaction` must be positive.')
            return None
        if self.j_local_exchange_interaction is None:
            self.j_local_exchange_interaction = 0.0 * ureg.eV
        return self.u_interaction - self.j_local_exchange_interaction

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)

        # Obtain (u, up, j_hunds_coupling) from slater_integrals
        if (
            self.u_interaction is None
            and self.u_interorbital_interaction is None
            and self.j_hunds_coupling is None
        ):
            (
                self.u_interaction,
                self.u_interorbital_interaction,
                self.j_hunds_coupling,
            ) = self.resolve_u_interactions(logger=logger)

        # If u_effective is not available, calculate it
        if self.u_effective is None:
            self.u_effective = self.resolve_u_effective(logger=logger)

        # Check if length of `orbitals_ref` is the same as the length of `umn`:
        if self.u_matrix is not None and self.orbitals_ref is not None:
            if len(self.u_matrix) != len(self.orbitals_ref):
                logger.error(
                    'The length of `HubbardInteractions.u_matrix` does not coincide with length of `HubbardInteractions.orbitals_ref`.'
                )


class AtomsState(Entity):
    """
    A base section to define each atom state information.
    """

    # TODO check what happens with ghost atoms that can have `chemical_symbol='X'`
    chemical_symbol = Quantity(
        type=MEnum(ase.data.chemical_symbols[1:]),
        description="""
        Symbol of the element, e.g. 'H', 'Pb'. This quantity is equivalent to `atomic_numbers`.
        """,
    )

    atomic_number = Quantity(
        type=np.int32,
        description="""
        Atomic number Z. This quantity is equivalent to `chemical_symbol`.
        """,
    )

    orbitals_state = SubSection(sub_section=OrbitalsState.m_def, repeats=True)

    charge = Quantity(
        type=np.int32,
        default=0,
        description="""
        Charge of the atom. It is defined as the number of extra electrons or holes in the
        atom. If the atom is neutral, charge = 0 and the summation of all (if available) the`OrbitalsState.occupation`
        coincides with the `atomic_number`. Otherwise, charge can be any positive integer (+1, +2...)
        for cations or any negative integer (-1, -2...) for anions.

        Note: for `CoreHole` systems we do not consider the charge of the atom even if
        we do not store the final `OrbitalsState` where the electron was excited to.
        """,
        a_eln=ELNAnnotation(component='NumberEditQuantity'),
    )

    core_hole = SubSection(sub_section=CoreHole.m_def, repeats=False)

    hubbard_interactions = SubSection(
        sub_section=HubbardInteractions.m_def, repeats=False
    )

    def resolve_chemical_symbol(self, logger: 'BoundLogger') -> Optional[str]:
        """
        Resolves the `chemical_symbol` from the `atomic_number`.

        Args:
            logger (BoundLogger): The logger to log messages.

        Returns:
            (Optional[str]): The resolved `chemical_symbol`.
        """
        if self.atomic_number is not None:
            try:
                return ase.data.chemical_symbols[self.atomic_number]
            except IndexError:
                logger.error(
                    'The `AtomsState.atomic_number` is out of range of the periodic table.'
                )
        return None

    def resolve_atomic_number(self, logger: 'BoundLogger') -> Optional[int]:
        """
        Resolves the `atomic_number` from the `chemical_symbol`.

        Args:
            logger (BoundLogger): The logger to log messages.

        Returns:
            (Optional[int]): The resolved `atomic_number`.
        """
        if self.chemical_symbol is not None:
            try:
                return ase.data.atomic_numbers[self.chemical_symbol]
            except IndexError:
                logger.error(
                    'The `AtomsState.chemical_symbol` is not recognized in the periodic table.'
                )
        return None

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)

        # Get chemical_symbol from atomic_number and viceversa
        if self.chemical_symbol is None:
            self.chemical_symbol = self.resolve_chemical_symbol(logger=logger)
        if self.atomic_number is None:
            self.atomic_number = self.resolve_atomic_number(logger=logger)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest
from typing import List, Optional

from nomad.datamodel import EntryArchive

from nomad_simulations.schema_packages.general import Simulation
from nomad_simulations.schema_packages.model_method import (
    TB,
    Wannier,
    SlaterKoster,
    SlaterKosterBond,
)
from nomad_simulations.schema_packages.atoms_state import OrbitalsState, AtomsState
from nomad_simulations.schema_packages.model_system import ModelSystem, AtomicCell

from . import logger
from .conftest import generate_simulation


class TestTB:
    """
    Test the `TB` class defined in `model_method.py`.
    """

    @pytest.mark.parametrize(
        'tb_section, result',
        [(Wannier(), 'Wannier'), (SlaterKoster(), 'SlaterKoster'), (TB(), None)],
    )
    def test_resolve_type(self, tb_section: TB, result: Optional[str]):
        """
        Test the `resolve_type` method.

        Args:
            tb_section (TB): The TB section to resolve the type from.
            result (Optional[str]): The expected type of the TB section.
        """
        assert tb_section.resolve_type() == result

    @pytest.mark.parametrize(
        'model_systems, model_index, result',
        [
            # no `ModelSystem` sections
            ([], 0, None),
            # `model_index` out of range
            ([ModelSystem()], 1, None),
            # no `is_representative` in `ModelSystem`
            ([ModelSystem(is_representative=False)], 0, None),
            # no `cell` section in `ModelSystem`
            ([ModelSystem(is_representative=True)], 0, None),
            # no `AtomsState` in `AtomicCell`
            ([ModelSystem(is_representative=True, cell=[AtomicCell()])], 0, None),
            # no `model_system` child section under `ModelSystem`
            (
                [
                    ModelSystem(
                        is_representative=True,
                        cell=[AtomicCell(atoms_state=[AtomsState()])],
                    )
                ],
                0,
                None,
            ),
            # `type` for the `model_system` child is not `'active_atom'`
            (
                [
                    ModelSystem(
                        is_representative=True,
                        cell=[AtomicCell(atoms_state=[AtomsState()])],
                        model_system=[ModelSystem(type='bulk')],
                    )
                ],
                0,
                [],
            ),
            # wrong index for `AtomsState in active atom
            (
                [
                    ModelSystem(
                        is_representative=True,
                        cell=[AtomicCell(atoms_state=[AtomsState()])],
                        model_system=[
                            ModelSystem(type='active_atom', atom_indices=[2])
                        ],
                    )
                ],
                0,
                [],
            ),
            # empty `OrbitalsState` in `AtomsState`
            (
                [
                    ModelSystem(
                        is_representative=True,
                        cell=[AtomicCell(atoms_state=[AtomsState(orbitals_state=[])])],
                        model_system=[
                            ModelSystem(type='active_atom', atom_indices=[0])
                        ],
                    )
                ],
                0,
                [],
            ),
            # valid case
            (
                [
                    ModelSystem(
                        is_representative=True,
                        cell=[
                            AtomicCell(
                                atoms_state=[
                                    AtomsState(
                                        orbitals_state=[
                                            OrbitalsState(l_quantum_symbol='s')
                                        ]
                                    )
                                ]
                            )
                        ],
                        model_system=[
                            ModelSystem(type='active_atom', atom_indices=[0])
                        ],
                    )
                ],
                0,
                [OrbitalsState(l_quantum_symbol='s')],
            ),
        ],
    )
    def test_resolve_orbital_references(
        self,
        model_systems: Optional[List[ModelSystem]],
        model_index: int,
        result: Optional[List[OrbitalsState]],
    ):
        """
        Test the `resolve_orbital_references` method.

        Args:
            model_systems (Optional[List[ModelSystem]]): The `model_system` section to add to `Simulation`.
            model_index (int): The index of the `ModelSystem` section to resolve the orbital references from.
            result (Optional[List[OrbitalsState]]): The expected orbital references.
        """
        tb_method = TB()
        simulation = generate_simulation(model_method=tb_method)
        simulation.model_system = model_systems
        orbitals_ref = tb_method.resolve_orbital_references(
            model_systems=model_systems,
            logger=logger,
            model_index=model_index,
        )
        if not orbitals_ref:
            assert orbitals_ref == result
        else:
            assert orbitals_ref[0].l_quantum_symbol == result[0].l_quantum_symbol

    @pytest.mark.parametrize(
        'tb_section, result_type, model_systems, result',
        [
            # no method `type` extracted
            (TB(), 'unavailable', [], None),
            # method `type` extracted
            (Wannier(), 'Wannier', [], None),
            # no `ModelSystem` sections
            (Wannier(), 'Wannier', [], None),
            # no `is_representative` in `ModelSystem`
            (Wannier(), 'Wannier', [ModelSystem(is_representative=False)], None),
            # no `cell` section in `ModelSystem`
            (Wannier(), 'Wannier', [ModelSystem(is_representative=True)], None),
            # no `AtomsState` in `AtomicCell`
            (
                Wannier(),
                'Wannier',
                [ModelSystem(is_representative=True, cell=[AtomicCell()])],
                None,
            ),
            # no `model_system` child section under `ModelSystem`
            (
                Wannier(),
                'Wannier',
                [
                    ModelSystem(
                        is_representative=True,
                        cell=[AtomicCell(atoms_state=[AtomsState()])],
                    )
                ],
                None,
            ),
            # `type` for the `model_system` child is not `'active_atom'`
            (
                Wannier(),
                'Wannier',
                [
                    ModelSystem(
                        is_representative=True,
                        cell=[AtomicCell(atoms_state=[AtomsState()])],
                        model_system=[ModelSystem(type='bulk')],
                    )
                ],
                None,
            ),
            # wrong index for `AtomsState in active atom
            (
                Wannier(),
                'Wannier',
                [
                    ModelSystem(
                        is_representative=True,
                        cell=[AtomicCell(atoms_state=[AtomsState()])],
                        model_system=[
                            ModelSystem(type='active_atom', atom_indices=[2])
                        ],
                    )
                ],
                None,
            ),
            # empty `OrbitalsState` in `AtomsState`
            (
                Wannier(),
                'Wannier',
                [
                    ModelSystem(
                        is_representative=True,
                        cell=[AtomicCell(atoms_state=[AtomsState(orbitals_state=[])])],
                        model_system=[
                            ModelSystem(type='active_atom', atom_indices=[0])
                        ],
                    )
                ],
                None,
            ),
            # `Wannier.orbitals_ref` already set up
            (
                Wannier(orbitals_ref=[OrbitalsState(l_quantum_symbol='d')]),
                'Wannier',
                [
                    ModelSystem(
                        is_representative=True,
                        cell=[
                            AtomicCell(
                                atoms_state=[
                                    AtomsState(
                                        orbitals_state=[
                                            OrbitalsState(l_quantum_symbol='s')
                                        ]
                                    )
                                ]
                            )
                        ],
                        model_system=[
                            ModelSystem(type='active_atom', atom_indices=[0])
                        ],
                    )
                ],
                [OrbitalsState(l_quantum_symbol='d')],
            ),
            # valid case
            (
                Wannier(),
                'Wannier',
                [
                    ModelSystem(
                        is_representative=True,
                        cell=[
                            AtomicCell(
                                atoms_state=[
                                    AtomsState(
                                        orbitals_state=[
                                            OrbitalsState(l_quantum_symbol='s')
                                        ]
                                    )
                                ]
                            )
                        ],
                        model_system=[
                            ModelSystem(type='active_atom', atom_indices=[0])
                        ],
                    )
                ],
                [OrbitalsState(l_quantum_symbol='s')],
            ),
        ],
    )
    def test_normalize(
        self,
        tb_section: TB,
        result_type: Optional[str],
        model_systems: Optional[List[ModelSystem]],
        result: Optional[List[OrbitalsState]],
    ):
        """
        Test the `resolve_orbital_references` method.

        Args:
            tb_section (TB): The TB section to resolve the type from.
            result_type (Optional[str]): The expected type of the TB section.
            model_systems (Optional[List[ModelSystem]]): The `model_system` section to add to `Simulation`.
            result (Optional[List[OrbitalsState]]): The expected orbital references.
        """
        simulation = generate_simulation(model_method=tb_section)
        simulation.model_system = model_systems
        tb_section.normalize(EntryArchive(), logger)
        assert tb_section.type == result_type
        if tb_section.orbitals_ref is not None:
            assert len(tb_section.orbitals_ref) == 1
            assert (
                tb_section.orbitals_ref[0].l_quantum_symbol
                == result[0].l_quantum_symbol
            )
        else:
            assert tb_section.orbitals_ref == result


class TestWannier:
    """
    Test the `Wannier` class defined in `model_method.py`.
    """

    @pytest.mark.parametrize(
        'localization_type, is_maximally_localized, result_localization_type',
        [
            # `localization_type` and `is_maximally_localized` are `None`
            (None, None, None),
            # `localization_type` set while `is_maximally_localized` is `None`
            ('single_shot', None, 'single_shot'),
            # normalizing from `is_maximally_localized`
            (None, True, 'maximally_localized'),
            (None, False, 'single_shot'),
        ],
    )
    def test_normalize(
        self,
        localization_type: Optional[str],
        is_maximally_localized: bool,
        result_localization_type: Optional[str],
    ):
        """
        Test the `normalize` method .

        Args:
            localization_type (Optional[str]): The localization type.
            is_maximally_localized (bool): If the localization is maximally-localized or a single-shot.
            result_localization_type (Optional[str]): The expected `localization_type` after normalization.
        """
        wannier = Wannier(
            localization_type=localization_type,
            is_maximally_localized=is_maximally_localized,
        )
        wannier.normalize(EntryArchive(), logger)
        assert wannier.localization_type == result_localization_type


class TestSlaterKosterBond:
    """
    Test the `SlaterKosterBond` class defined in `model_method.py`.
    """

    @pytest.mark.parametrize(
        'orbital_1, orbital_2, bravais_vector, result',
        [
            # no `OrbitalsState` sections
            (None, None, (), None),
            (None, OrbitalsState(), (), None),
            (OrbitalsState(), None, (), None),
            # no `bravais_vector`
            (OrbitalsState(), OrbitalsState(), None, None),
            # no `l_quantum_symbol` in `OrbitalsState`
            (OrbitalsState(), OrbitalsState(), (0, 0, 0), None),
            # valid cases
            (
                OrbitalsState(l_quantum_symbol='s'),
                OrbitalsState(l_quantum_symbol='s'),
                (0, 0, 0),
                'sss',
            ),
            (
                OrbitalsState(l_quantum_symbol='s'),
                OrbitalsState(l_quantum_symbol='p'),
                (0, 0, 0),
                'sps',
            ),
        ],
    )
    def test_resolve_bond_name_from_references(
        self,
        orbital_1: Optional[OrbitalsState],
        orbital_2: Optional[OrbitalsState],
        bravais_vector: Optional[tuple],
        result: Optional[str],
    ):
        """
        Test the `resolve_bond_name_from_references` method.

        Args:
            orbital_1 (Optional[OrbitalsState]): The first `OrbitalsState` section.
            orbital_2 (Optional[OrbitalsState]): The second `OrbitalsState` section.
            bravais_vector (Optional[tuple]): The bravais vector.
            result (Optional[str]): The expected bond name.
        """
        sk_bond = SlaterKosterBond()
        bond_name = sk_bond.resolve_bond_name_from_references(
            orbital_1=orbital_1,
            orbital_2=orbital_2,
            bravais_vector=bravais_vector,
            logger=logger,
        )
        assert bond_name == result

    @pytest.mark.parametrize(
        'orbital_1, orbital_2, bravais_vector, result',
        [
            # no `OrbitalsState` sections
            (None, None, [], None),
            (None, OrbitalsState(), [], None),
            (OrbitalsState(), None, [], None),
            # no `bravais_vector`
            (OrbitalsState(), OrbitalsState(), None, None),
            # no `l_quantum_symbol` in `OrbitalsState`
            (OrbitalsState(), OrbitalsState(), (0, 0, 0), None),
            # valid cases
            (
                OrbitalsState(l_quantum_symbol='s'),
                OrbitalsState(l_quantum_symbol='s'),
                (0, 0, 0),
                'sss',
            ),
            (
                OrbitalsState(l_quantum_symbol='s'),
                OrbitalsState(l_quantum_symbol='p'),
                (0, 0, 0),
                'sps',
            ),
        ],
    )
    def test_normalize(
        self,
        orbital_1: Optional[OrbitalsState],
        orbital_2: Optional[OrbitalsState],
        bravais_vector: Optional[tuple],
        result: Optional[str],
    ):
        """
        Test the `normalize` method.

        Args:
            orbital_1 (Optional[OrbitalsState]): The first `OrbitalsState` section.
            orbital_2 (Optional[OrbitalsState]): The second `OrbitalsState` section.
            bravais_vector (Optional[tuple]): The bravais vector.
            result (Optional[str]): The expected SK bond `name` after normalization.
        """
        sk_bond = SlaterKosterBond()
        atoms_state = AtomsState()
        simulation = Simulation(
            model_system=[ModelSystem(cell=[AtomicCell(atoms_state=[atoms_state])])]
        )
        if orbital_1 is not None:
            atoms_state.orbitals_state.append(orbital_1)
            sk_bond.orbital_1 = atoms_state.orbitals_state[0]
        if orbital_2 is not None:
            atoms_state.orbitals_state.append(orbital_2)
            sk_bond.orbital_2 = atoms_state.orbitals_state[-1]
        if bravais_vector is not None and len(bravais_vector) != 0:
            sk_bond.bravais_vector = bravais_vector
        sk_bond.normalize(EntryArchive(), logger)
        assert sk_bond.name == result



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest
import numpy as np
from typing import List, Optional

from nomad.units import ureg
from nomad.datamodel import EntryArchive

from nomad_simulations.schema_packages.general import Simulation
from nomad_simulations.schema_packages.model_system import ModelSystem, AtomicCell
from nomad_simulations.schema_packages.atoms_state import AtomsState
from nomad_simulations.schema_packages.outputs import Outputs
from nomad_simulations.schema_packages.properties import (
    SpectralProfile,
    ElectronicDensityOfStates,
    AbsorptionSpectrum,
    XASSpectrum,
)
from nomad_simulations.schema_packages.variables import Energy2 as Energy

from . import logger


class TestSpectralProfile:
    """
    Test the `SpectralProfile` class defined in `properties/spectral_profile.py`.
    """

    def test_is_valid_spectral_profile(self):
        """
        Test the `is_valid_spectral_profile` method.
        """
        spectral_profile = SpectralProfile(
            variables=[Energy(points=[-1, 0, 1] * ureg.joule)]
        )
        spectral_profile.value = [1.5, 0, 0.8]
        assert spectral_profile.is_valid_spectral_profile()
        spectral_profile.value = [2, 0, -4]
        assert not spectral_profile.is_valid_spectral_profile()


class TestElectronicDensityOfStates:
    """
    Test the `ElectronicDensityOfStates` class defined in `properties/spectral_profile.py`.
    """

    # ! Include this initial `test_default_quantities` method when testing your PhysicalProperty classes
    def test_default_quantities(self):
        """
        Test the default quantities assigned when creating an instance of the `ElectronicDensityOfStates` class.
        """
        electronic_dos = ElectronicDensityOfStates()
        assert (
            electronic_dos.iri
            == 'http://fairmat-nfdi.eu/taxonomy/ElectronicDensityOfStates'
        )
        assert electronic_dos.name == 'ElectronicDensityOfStates'
        assert electronic_dos.rank == []

    def test_resolve_energies_origin(self):
        """
        Test the `resolve_energies_origin` method.
        """
        # ! add test when `ElectronicEigenvalues` is implemented
        pass

    def test_resolve_normalization_factor(self, simulation_electronic_dos: Simulation):
        """
        Test the `resolve_normalization_factor` method.
        """
        simulation = Simulation()
        outputs = Outputs()
        # We only used the `simulation_electronic_dos` fixture to get the `ElectronicDensityOfStates` to test missing refs
        electronic_dos = simulation_electronic_dos.outputs[0].electronic_dos[0]
        electronic_dos.energies_origin = 0.5 * ureg.joule
        outputs.electronic_dos.append(electronic_dos)
        simulation.outputs.append(outputs)

        # No `model_system_ref`
        assert electronic_dos.resolve_normalization_factor(logger) is None

        # No `model_system_ref.cell`
        model_system = ModelSystem()
        simulation.model_system.append(model_system)
        outputs.model_system_ref = simulation.model_system[0]
        assert electronic_dos.resolve_normalization_factor(logger) is None

        # No `model_system_ref.cell.atoms_state`
        atomic_cell = AtomicCell(
            type='original', positions=[[0, 0, 0], [0.5, 0.5, 0.5]] * ureg.meter
        )
        model_system.cell.append(atomic_cell)
        assert electronic_dos.resolve_normalization_factor(logger) is None

        # Adding the required `model_system_ref` sections and quantities
        atoms_state = [
            AtomsState(chemical_symbol='Ga'),
            AtomsState(chemical_symbol='As'),
        ]
        for atom in atoms_state:
            atom.normalize(EntryArchive(), logger)
        atomic_cell.atoms_state = atoms_state
        # Non spin-polarized
        normalization_factor = electronic_dos.resolve_normalization_factor(logger)
        assert np.isclose(normalization_factor, 0.015625)
        # Spin-polarized
        electronic_dos.spin_channel = 0
        normalization_factor_spin_polarized = (
            electronic_dos.resolve_normalization_factor(logger)
        )
        assert np.isclose(
            normalization_factor_spin_polarized, 0.5 * normalization_factor
        )

    def test_extract_band_gap(self):
        """
        Test the `extract_band_gap` method.
        """
        # ! add test when `ElectronicEigenvalues` is implemented
        pass

    def test_resolve_pdos_name(self, simulation_electronic_dos: Simulation):
        """
        Test the `resolve_pdos_name` method.
        """
        # Get projected DOSProfile from the simulation fixture
        projected_dos = (
            simulation_electronic_dos.outputs[0].electronic_dos[0].projected_dos
        )
        assert len(projected_dos) == 3
        pdos_names = ['orbital s Ga', 'orbital px As', 'orbital py As']
        for i, pdos in enumerate(projected_dos):
            name = pdos.resolve_pdos_name(logger)
            assert name == pdos_names[i]

    def test_extract_projected_dos(self, simulation_electronic_dos: Simulation):
        """
        Test the `extract_projected_dos` method.
        """
        # Get Outputs and ElectronicDensityOfStates from the simulation fixture
        outputs = simulation_electronic_dos.outputs[0]
        electronic_dos = outputs.electronic_dos[0]

        # Initial tests for the passed `projected_dos` (only orbital PDOS)
        assert len(electronic_dos.projected_dos) == 3  # only orbital projected DOS
        orbital_projected = electronic_dos.extract_projected_dos('orbital', logger)
        atom_projected = electronic_dos.extract_projected_dos('atom', logger)
        assert len(orbital_projected) == 3 and len(atom_projected) == 0
        orbital_projected_names = [orb_pdos.name for orb_pdos in orbital_projected]
        assert orbital_projected_names == [
            'orbital s Ga',
            'orbital px As',
            'orbital py As',
        ]
        assert (
            orbital_projected[0].entity_ref
            == outputs.model_system_ref.cell[0].atoms_state[0].orbitals_state[0]
        )  # orbital `s` in `Ga` atom
        assert (
            orbital_projected[1].entity_ref
            == outputs.model_system_ref.cell[0].atoms_state[1].orbitals_state[0]
        )  # orbital `px` in `As` atom
        assert (
            orbital_projected[2].entity_ref
            == outputs.model_system_ref.cell[0].atoms_state[1].orbitals_state[1]
        )  # orbital `py` in `As` atom

        orbital_projected = electronic_dos.extract_projected_dos('orbital', logger)
        atom_projected = electronic_dos.extract_projected_dos('atom', logger)
        assert len(orbital_projected) == 3 and len(atom_projected) == 0

    @pytest.mark.parametrize(
        'value, result',
        [
            (None, [1.5, 1.2, 0, 0, 0, 0.8, 1.3]),
            ([30.5, 1.2, 0, 0, 0, 0.8, 1.3], [30.5, 1.2, 0, 0, 0, 0.8, 1.3]),
        ],
    )
    def test_generate_from_pdos(
        self,
        simulation_electronic_dos: Simulation,
        value: Optional[List[float]],
        result: List[float],
    ):
        """
        Test the `generate_from_projected_dos` method.
        """
        # Get Outputs and ElectronicDensityOfStates from the simulation fixture
        outputs = simulation_electronic_dos.outputs[0]
        electronic_dos = outputs.electronic_dos[0]

        # Add `value`
        if value is not None:
            electronic_dos.value = value * ureg('1/joule')

        val = electronic_dos.generate_from_projected_dos(logger)
        assert (val.magnitude == result).all()

        # Testing orbital + atom projected DOS (3 orbitals + 2 atoms PDOS)
        assert len(electronic_dos.projected_dos) == 5
        orbital_projected = electronic_dos.extract_projected_dos('orbital', logger)
        atom_projected = electronic_dos.extract_projected_dos('atom', logger)
        assert len(orbital_projected) == 3 and len(atom_projected) == 2
        atom_projected_names = [atom_pdos.name for atom_pdos in atom_projected]
        assert atom_projected_names == ['atom Ga', 'atom As']
        assert (
            atom_projected[0].entity_ref
            == outputs.model_system_ref.cell[0].atoms_state[0]
        )  # `Ga` atom
        assert (
            atom_projected[1].entity_ref
            == outputs.model_system_ref.cell[0].atoms_state[1]
        )  # `As` atom

    def test_normalize(self):
        """
        Test the `normalize` method.
        """
        # ! add test when `ElectronicEigenvalues` is implemented
        pass


class TestAbsorptionSpectrum:
    """
    Test the `AbsorptionSpectrum` class defined in `properties/spectral_profile.py`.
    """

    # ! Include this initial `test_default_quantities` method when testing your PhysicalProperty classes
    def test_default_quantities(self):
        """
        Test the default quantities assigned when creating an instance of the `AbsorptionSpectrum` class.
        """
        absorption_spectrum = AbsorptionSpectrum()
        assert absorption_spectrum.iri is None  # Add iri when available
        assert absorption_spectrum.name == 'AbsorptionSpectrum'
        assert absorption_spectrum.rank == []


class TestXASSpectrum:
    """
    Test the `XASSpectrum` class defined in `properties/spectral_profile.py`.
    """

    # ! Include this initial `test_default_quantities` method when testing your PhysicalProperty classes
    def test_default_quantities(self):
        """
        Test the default quantities assigned when creating an instance of the `XASSpectrum` class.
        """
        xas_spectrum = XASSpectrum()
        assert xas_spectrum.iri is None  # Add iri when available
        assert xas_spectrum.name == 'XASSpectrum'
        assert xas_spectrum.rank == []

    @pytest.mark.parametrize(
        'xanes_energies, exafs_energies, xas_values',
        [
            (None, None, None),
            ([0, 1, 2], None, None),
            (None, [3, 4, 5], None),
            ([0, 1, 2], [3, 4, 5], [0.5, 0.1, 0.3, 0.2, 0.4, 0.6]),
            ([0, 1, 4], [3, 4, 5], None),
            ([0, 1, 2], [0, 4, 5], None),
        ],
    )
    def test_generate_from_contributions(
        self,
        xanes_energies: Optional[List[float]],
        exafs_energies: Optional[List[float]],
        xas_values: Optional[List[float]],
    ):
        """
        Test the `generate_from_contributions` method.
        """
        xas_spectrum = XASSpectrum()
        if xanes_energies is not None:
            xanes_spectrum = AbsorptionSpectrum()
            xanes_spectrum.variables = [Energy(points=xanes_energies * ureg.joule)]
            xanes_spectrum.value = [0.5, 0.1, 0.3]
            xas_spectrum.xanes_spectrum = xanes_spectrum
        if exafs_energies is not None:
            exafs_spectrum = AbsorptionSpectrum()
            exafs_spectrum.variables = [Energy(points=exafs_energies * ureg.joule)]
            exafs_spectrum.value = [0.2, 0.4, 0.6]
            xas_spectrum.exafs_spectrum = exafs_spectrum
        xas_spectrum.generate_from_contributions(logger)
        if xas_spectrum.value is not None:
            assert (xas_spectrum.value == xas_values).all()
        else:
            assert xas_spectrum.value == xas_values
